#! /usr/bin/env python
#
# GOAL
#   - load in lnL data
#   - fit peak to quadratic (standard), GP, etc. 
#   - pass as input to mcsampler, to generate posterior samples
#
# FORMAT
#   - pankow simplification of standard format
#
# COMPARE TO
#   util_NRQuadraticFit.py
#   postprocess_1d_cumulative
#   util_QuadraticMassPosterior.py
#



import argparse
import sys
import numpy as np
import numpy.lib.recfunctions
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import functools
import itertools


# GPU acceleration: NOT YET, just do usual
xpy_default=numpy  # just in case, to make replacement clear and to enable override
identity_convert = lambda x: x  # trivial return itself
cupy_success=False


no_plots = True
internal_dtype = np.float32  # only use 32 bit storage! Factor of 2 memory savings for GP code in high dimensions

C_CGS=2.997925*10**10 # Argh, Monica!
 


parser = argparse.ArgumentParser()
parser.add_argument("--fname",help="filename of *.dat file [standard ILE output]")
parser.add_argument("--input-tides",action='store_true',help="Use input format with tidal fields included.")
parser.add_argument("--input-distance",action='store_true',help="Use input format with distance fields (but not tidal fields?) enabled.")
parser.add_argument("--fname-lalinference",help="filename of posterior_samples.dat file [standard LI output], to overlay on corner plots")
parser.add_argument("--fname-output-samples",default="output-ILE-samples",help="output posterior samples (default output-ILE-samples -> output-ILE)")
parser.add_argument("--fname-output-integral",default="integral_result",help="output filename for integral result. Postfixes appended")
parser.add_argument("--approx-output",default="SEOBNRv2", help="approximant to use when writing output XML files.")
parser.add_argument("--amplitude-order",default=-1,type=int,help="Set ampO for grid. Used in PN")
parser.add_argument("--phase-order",default=7,type=int,help="Set phaseO for grid. Used in PN")
parser.add_argument("--fref",default=20,type=float, help="Reference frequency used for spins in the ILE output.  (Since I usually use SEOBNRv3, the best choice is 20Hz)")
parser.add_argument("--fmin",type=float,default=20)
parser.add_argument("--fname-rom-samples",default=None,help="*.rom_composite output. Treated identically to set of posterior samples produced by mcsampler after constructing fit.")
parser.add_argument("--n-output-samples",default=3000,type=int,help="output posterior samples (default 3000)")
parser.add_argument("--desc-lalinference",type=str,default='',help="String to adjoin to legends for LI")
parser.add_argument("--desc-ILE",type=str,default='',help="String to adjoin to legends for ILE")
parser.add_argument("--parameter", action='append', help="Parameters used as fitting parameters AND varied at a low level to make a posterior")
parser.add_argument("--parameter-implied", action='append', help="Parameter used in fit, but not independently varied for Monte Carlo")
parser.add_argument("--no-adapt-parameter",action='append',help="Disable adaptive sampling in a parameter. Useful in cases where a parameter is not well-constrained, and the a prior sampler is well-chosen.")
parser.add_argument("--mc-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--eta-range",default=None,help="Eta range. Important if we have a BNS or other item that has a strong constraint.")
parser.add_argument("--mtot-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--trust-sample-parameter-box",action='store_true', help="If used, sets the prior range to the SAMPLE range for any parameters. NOT IMPLEMENTED. This should be automatically done for mc!")
parser.add_argument("--plots-do-not-force-large-range",action='store_true', help = "If used, the plots do NOT automatically set the chieff range to [-1,1], the eta range to [0,1/4], etc")
parser.add_argument("--downselect-parameter",action='append', help='Name of parameter to be used to eliminate grid points ')
parser.add_argument("--downselect-parameter-range",action='append',type=str)
parser.add_argument("--no-downselect",action='store_true',help='Prevent using downselection on output points' )
parser.add_argument("--no-downselect-grid",action='store_true',help='Prevent using downselection on input points. Applied only to mc range' )
parser.add_argument("--downselect-enforce-kerr",action='store_true',help="Provides limits that enforce the kerr limit. Also imposed in coordinate transformations.")
parser.add_argument("--aligned-prior", default="uniform",help="Options are 'uniform', 'volumetric', and 'alignedspin-zprior'. Only influences s1z, s2z")
parser.add_argument("--transverse-prior", default="uniform",help="Options are 'volumetric' (default) and 'alignedspin-zprior'. Only influences s1x,s1y,s2x,s2y")
parser.add_argument("--spin-prior-chizplusminus-alternate-sampling",default='alignedspin_zprior',help="Use gaussian sampling when using chizplus, chizminus, to make reweighting more efficient.")
parser.add_argument("--import-prior-dictionary-file",default=None,type=str,help="File with dictionary stored_param_dict = 'name':func and stored_param_ranges = 'name':[left,right].  Use to overwrite priors with user-specified function")
parser.add_argument("--output-prior-dictionary-file",default=None,type=str,help="File with dictionary 'name':func. ")
parser.add_argument("--prior-gaussian-mass-ratio",action='store_true',help="Applies a gaussian mass ratio prior (mean=0.5, width=0.2 by default). Only viable in mtot, q coordinates. Not properly normalized, so will break bayes factors by about 2%%")
parser.add_argument("--prior-tapered-mass-ratio",action='store_true',help="Applies a tapered mass ratio prior (transition 0.8, kappa=20). Only viable in mtot, q coordinates. Not properly normalized, a tapering factor instread")
parser.add_argument("--prior-gaussian-spin1-magnitude",action='store_true',help="Applies a gaussian spin magnitude prior (mean=0.7, width=0.1 by default) for FIRST spin. Only viable in polar spin coordinates. Not properly normalized, so will break bayes factors by a small amount (depending on chi_max).  Used for 2g+1g merger arguments")
parser.add_argument("--prior-tapered-spin1-magnitude",action='store_true',help="Applies a tapered prior to spin1 magnitude")
parser.add_argument("--prior-tapered-spin1z",action='store_true',help="Applies a tapered prior to spin1's z component")
parser.add_argument("--pseudo-uniform-magnitude-prior", action='store_true',help="Applies volumetric prior internally, and then reweights at end step to get uniform spin magnitude prior")
parser.add_argument("--pseudo-uniform-magnitude-prior-alternate-sampling", action='store_true',help="Changes the internal sampling to be gaussian, not volumetric")
parser.add_argument("--pseudo-gaussian-mass-prior",action='store_true', help="Applies a gaussian mass prior in postprocessing. Done via reweighting so we can use arbitrary mass sampling coordinates.")
parser.add_argument("--pseudo-gaussian-mass-prior-mean",default=1.33,type=float, help="Mean value for reweighting")
parser.add_argument("--pseudo-gaussian-mass-prior-std",default=0.09, type=float,help="Width for reweighting")
parser.add_argument("--prior-lambda-linear",action='store_true',help="Use p(lambda) ~ lambdamax -lambda. Intended for first few iterations, to insure strong coverage of the low-lambda_k corner")
parser.add_argument("--prior-lambda-power",type=float,default=1,help="Use p(lambda) ~ (lambdamax -lambda)^p. Intended for first few iterations, to insure strong coverage of the low-lambda_k corner")
parser.add_argument("--mirror-points",action='store_true',help="Use if you have many points very near equal mass (BNS). Doubles the number of points in the fit, each of which has a swapped m1,m2")
parser.add_argument("--cap-points",default=-1,type=int,help="Maximum number of points in the sample, if positive. Useful to cap the number of points ued for GP. See also lnLoffset. Note points are selected AT RANDOM")
parser.add_argument("--chi-max", default=1,type=float,help="Maximum range of 'a' allowed.  Use when comparing to models that aren't calibrated to go to the Kerr limit.")
parser.add_argument("--chi-small-max", default=None,type=float,help="Maximum range of 'a' allowed on the smaller body.  If not specified, defaults to chi_max")
parser.add_argument("--chiz-plus-range", default=None,help="USE WITH CARE: If you are using chiz_minus, chiz_plus for a near-equal-mass system, then setting the chiz-plus-range can improve convergence (e.g., for aligned-spin systems), loosely by setting a chi_eff range that is allowed")
parser.add_argument("--lambda-max", default=4000,type=float,help="Maximum range of 'Lambda' allowed.  Minimum value is ZERO, not negative.")
parser.add_argument("--lambda-small-max", default=None,type=float,help="Maximum range of 'Lambda' allowed for smaller body. If provided and smaller than lambda_max, used ")
parser.add_argument("--lambda-plus-max", default=None,type=float,help="Maximum range of 'Lambda_plus' allowed.  Used for sampling. Pick small values to accelerate sampling! Otherwise, use lambda-max.")
parser.add_argument("--parameter-nofit", action='append', help="Parameter used to initialize the implied parameters, and varied at a low level, but NOT the fitting parameters")
parser.add_argument("--use-precessing",action='store_true')
parser.add_argument("--lnL-shift-prevent-overflow",default=None,type=float,help="Define this quantity to be a large positive number to avoid overflows. Note that we do *not* define this dynamically based on sample values, to insure reproducibility and comparable integral results. BEWARE: If you shift the result to be below zero, because the GP relaxes to 0, you will get crazy answers.")
parser.add_argument("--lnL-protect-overflow",action='store_true',help="Before fitting, subtract lnLmax - 100.  Add this quantity back at the end.")
parser.add_argument("--lnL-offset",type=float,default=np.inf,help="lnL offset. ONLY POINTS within lnLmax - lnLoffset are used in the calculation!  VERY IMPORTANT - default value chosen to include all points, not viable for production with some fit techniques like gp")
parser.add_argument("--lnL-offset-n-random",type=int,default=0,help="Add this many random points past the threshold")
parser.add_argument("--lnL-cut",type=float,default=None,help="lnL cut [MANUAL]. Remove points below this likelihood value from consideration.  Generally should not use")
parser.add_argument("--M-max-cut",type=float,default=1e5,help="Maximum mass to consider (e.g., if there is a cut on distance, this matters)")
parser.add_argument("--sigma-cut",type=float,default=0.6,help="Eliminate points with large error from the fit.")
parser.add_argument("--ignore-errors-in-data",action='store_true',help='Ignore reported error in lnL. Helpful for testing purposes (i.e., if the error is zero)')
parser.add_argument("--lnL-peak-insane-cut",type=float,default=np.inf,help="Throw away lnL greater than this value. Should not be necessary")
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--save-plots",default=False,action='store_true', help="Write plots to file (only useful for OSX, where interactive is default")
parser.add_argument("--inj-file", help="Name of injection file")
parser.add_argument("--event-num", type=int, default=0,help="Zero index of event in inj_file")
parser.add_argument("--report-best-point",action='store_true')
parser.add_argument("--force-no-adapt",action='store_true',help="Disable adaptation, both of the tempering exponent *and* the individual sampling prior(s)")
parser.add_argument("--fit-uses-reported-error",action='store_true')
parser.add_argument("--fit-uses-reported-error-factor",type=float,default=1,help="Factor to add to standard deviation of fit, before adding to lnL. Multiplies number fitting dimensions")
parser.add_argument("--n-max",default=3e8,type=float)
parser.add_argument("--n-eff",default=3e3,type=int)
parser.add_argument("--internal-bound-factor-if-n-eff-small",default=None,type=float,help="If n_eff < n_ouptut_samples, we truncate the output size based on n_eff*(factor)")
parser.add_argument("--n-chunk",default=1e5,type=int)
parser.add_argument("--contingency-unevolved-neff",default=None,help="Contingency planning for when n_eff produced by CIP is small, and user doesn't want to have hard failures.  Note --fail-unless-n-eff will prevent this from happening. Options: quadpuff, ...")
parser.add_argument("--not-worker",action='store_true',help="Nonworker jobs, IF we have workers present, don't have the 'fail unless' statement active")
parser.add_argument("--fail-unless-n-eff",default=None,type=int,help="If nonzero, places a minimum requirement on n_eff. Code will exit if not achieved, with no sample generation")
parser.add_argument("--fit-method",default="quadratic",help="quadratic|polynomial|gp|gp_hyper|gp_lazy|cov|kde")
parser.add_argument("--fit-load-quadratic",default=None,help="Filename of hdf5 file to load quadratic fit from. ")
parser.add_argument("--fit-load-quadratic-path",default="GW190814/annealing_mc_source_eta_chieff",help="Path in hdf5 file to specific covariance matrix to be used")
parser.add_argument("--pool-size",default=3,type=int,help="Integer. Number of GPs to use (result is averaged)")
parser.add_argument("--fit-load-gp",default=None,type=str,help="Filename of GP fit to load. Overrides fitting process, but user MUST correctly specify coordinate system to interpret the fit with.  Does not override loading and converting the data.")
parser.add_argument("--fit-save-gp",default=None,type=str,help="Filename of GP fit to save. ")
parser.add_argument("--fit-order",type=int,default=2,help="Fit order (polynomial case: degree)")
parser.add_argument("--fit-uncertainty-added",default=False, action='store_true', help="Reported likelihood is lnL+(fit error). Use for placement and use of systematic errors.")
parser.add_argument("--no-plots",action='store_true')
parser.add_argument("--using-eos", type=str, default=None, help="Name of EOS.  Fit parameter list should physically use lambda1, lambda2 information (but need not) ")
parser.add_argument("--no-use-lal-eos",action='store_true',help="Do not use LAL EOS interface. Used for spectral EOS. Do not use this.")
parser.add_argument("--no-matter1", action='store_true', help="Set the lambda parameters to zero (BBH) but return them")
parser.add_argument("--no-matter2", action='store_true', help="Set the lambda parameters to zero (BBH) but return them")
parser.add_argument("--protect-coordinate-conversions", action='store_true', help="Adds an extra layer to coordinate conversions with range tests. Slows code down, but adds layer of safety for out-of-range EOS parameters for example")
parser.add_argument("--source-redshift",default=0,type=float,help="Source redshift (used to convert from source-frame mass [integration limits] to arguments of fitting function.  Note that if nonzero, integration done in SOURCE FRAME MASSES, but the fit is calculated using DETECTOR FRAME")
parser.add_argument("--eos-param", type=str, default=None, help="parameterization of equation of state")
parser.add_argument("--eos-param-values", default=None, help="Specific parameter list for EOS")
parser.add_argument("--sampler-method",default="adaptive_cartesian",help="adaptive_cartesian|GMM|adaptive_cartesian_gpu")
parser.add_argument("--internal-use-lnL",action='store_true',help="integrator internally manipulates lnL. ONLY VIABLE FOR GMM AT PRESENT")
parser.add_argument("--internal-correlate-parameters",default=None,type=str,help="comman-separated string indicating parameters that should be sampled allowing for correlations. Must be sampling parameters. Only implemented for gmm.  If string is 'all', correlate *all* parameters")
parser.add_argument("--internal-n-comp",default=1,type=int,help="number of components to use for GMM sampling. Default is 1, because we expect a unimodal posterior in well-adapted coordinates.  If you have crappy coordinates, use more")
parser.add_argument("--internal-gmm-memory-chisquared-factor",default=None,type=float,help="Multiple of the number of degrees of freedom to save. 5 is a part in 10^6, 4 is 10^{-4}, and None keeps all up to lnL_offset.  Note that low-weight points can contribute notably to n_eff, and it can be dangerous to assume a simple chisquared likelihood!  Provided in case we need very long runs")
parser.add_argument("--use-eccentricity", action="store_true")

ECC_MAX = 0.25 # maximum value of eccentricity, hard-coding here for ease of editing

# FIXME hacky options added by me (Liz) to try to get my capstone project to work.
# I needed a way to fix the component masses and nothing else seemed to work.
parser.add_argument("--fixed-parameter", action="append")
parser.add_argument("--fixed-parameter-value", action="append")

opts=  parser.parse_args()
if not(opts.no_adapt_parameter):
    opts.no_adapt_parameter =[] # needs to default to empty list
no_plots = no_plots |  opts.no_plots
lnL_shift = 0
lnL_default_large_negative = -500
if opts.lnL_shift_prevent_overflow:
    lnL_shift  = opts.lnL_shift_prevent_overflow
if not(opts.force_no_adapt):
    opts.force_no_adapt=False  # force explicit boolean false

print(" Arguments: ",' '.join(sys.argv[1:]))
print(" Input: ",opts.fname)
print(" Output: ",opts.fname_output_samples+".xml.gz")

import os
os.system("touch "+opts.fname_output_samples + ".xml.gz")

print(" ==  Input contents (at execute time) == ")
os.system("cat {} ".format(opts.fname))


os.system("wc -l {} > {} ".format(opts.fname,opts.fname_output_samples + ".xml.gz")) # put input contents into the output file, so I can see the inputs




#! /usr/bin/env python
#
# Attempt at semi-automated PP plot.  Shows empirical CDF (p value vs estimate), and 90% credible interval of *most extreme* of the N things being plotted
# 
# FIXME
#   - strip data without convergence test
#   - enable column names 
#
# EXAMPLE
#    python pp_plot.py net_pp.dat_clean 2 "['mc','q']"
#   python pp_plot.py net_pp.dat_clean 4 "['{\cal M}_c','q','\chi_{1,z}','\chi_{2,z}']"


import numpy as np
from matplotlib import pyplot as plt
from scipy.stats import norm
import scipy.stats

dpi_base=200
legend_font_base=16
rc_params = {'backend': 'ps',
             'axes.labelsize': 11,
             'axes.titlesize': 10,
             'font.size': 11,
             'legend.fontsize': legend_font_base,
             'xtick.labelsize': 11,
             'ytick.labelsize': 11,
             #'text.usetex': True,
             'font.family': 'Times New Roman'}#,
             #'font.sans-serif': ['Bitstream Vera Sans']}#,
plt.rcParams.update(rc_params)


def binomial_credible_interval(phat,n,z):
    offset = np.sqrt(phat*(1-phat)/n + z**2/(4*n**2))
    my_1 = np.ones((len(phat),2))
    return          1./(1 + z**2/n) *  (np.outer(phat,np.array([1,1])) +  my_1* z**2/(2* n) + np.outer(z *offset, np.array([-1,1])))
    
pvalFiducial = 0.9 # fiducial credible level for outermost CI.  (Will change depending on number of parameters)

def binomial_credible_interval_default(phat,n,nParams=2):
    z_fiducial = norm.ppf((  np.power(1-(1.-pvalFiducial)/2, 1./nParams)))
    print("fiducial z", z_fiducial, " for nParams = ",nParams)
    return  binomial_credible_interval(phat,n, z_fiducial)

import sys

dat = np.genfromtxt(sys.argv[1],filling_values=1e9,invalid_raise=False,usecols=tuple(np.arange(int(sys.argv[2]))))
if False:
    # Stripping invalid entries: only works if last column *used* is correctly loaded, and we retain all output parameters
    len_orig = len(dat)
    print(dat[:,-1])
    dat = dat[ dat[:,-1]<1e-2]
    print(" Reducing size from ", len_orig, " to ", len(dat))
if len(sys.argv) > 3:
    param_labels=eval(sys.argv[3])#sys.argv[3].split()  # assume a string is passed, with whitespace
else:
    param_labels = np.arange(int(sys.argv[2]))

nParams = len(dat[0])
for indx in np.arange(nParams):
    pvals = np.sort(dat[:,indx])
    pvals_emp = np.arange(len(dat))*1.0/len(dat) 
    plt.scatter(pvals,pvals_emp,label='$'+str(param_labels[indx])+'$')
    # KS test, single instance
    print(" KS {} ".format(param_labels[indx]), scipy.stats.kstest(pvals,'uniform'))
    

xvals = np.linspace(0,1,100)
plt.plot(xvals,xvals,color='k')
pvals_lims = binomial_credible_interval_default(xvals, len(dat),nParams=nParams)
plt.plot(xvals,pvals_lims[:,0], color='k',ls=':')
plt.plot(xvals,pvals_lims[:,1], color='k',ls=':')
plt.legend()
plt.xlabel(r"$P(x_{\rm inj})$")
plt.ylabel(r"$\hat{P}$")
plt.savefig("prior_pp.pdf",dpi=dpi_base)



#! /usr/bin/env python
# Objective
#    call bilby  make_pp_plot on RIFT results : use same formatting, error bars, and overall p-value
# Compare to
#     https://git.ligo.org/pe/O4/bilby_o4_review/-/blob/main/pp-test-templates/make_pp_plot.py


import glob
import json
import sys
import argparse

import lal
import lalsimulation as lalsim

from collections import namedtuple

import bilby
import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import RIFT.lalsimutils as lalsimutils
from RIFT.misc.samples_utils import add_field

from itertools import product
import scipy.stats
from bilby.core.utils import logger
from bilby.core.utils import safe_save_figure

param_names = ['mc','q', 'ra', 'dec','incl', 'psi', 'phiorb','distance']
remap_export_to_P = {'mc':'mc', 'q':'q','ra':'phi', 'dec':'theta', 'distance':'dist','phiorb':'phiref','psi':'psi','incl':'incl'}
for i in ['1','2']:
    for j in ['x','y','z']:
        str_in = 's'+i+j
        str_out = 'a'+i+j
        remap_export_to_P[str_out]=str_in
# Omit phases, they are too hard to get right in a quick pass
param_names_alt = ['mc','q','chi1', 'chi2', 'theta1','theta2', 'thetaJN','phiorb', 'distance','psi'] # 'psiJ', 'phiJL', 'phi12'

for name in ['chi1', 'chi2', 'theta1','theta2', 'thetaJN', 'psiJ', 'phiJL', 'phi12']:
   remap_export_to_P[name]=name


fref=20

def load_result(fname,alt_coords=False):
#    print(fname)
    samples = np.genfromtxt(fname,names=True)
    samples['psi'] = np.mod(samples['psi'],np.pi) # deal with range issue
    # Coordinate conversion
    if alt_coords:
        samples = add_field(samples, [('chi1',float),('chi2',float), ('theta1',float), ('theta2',float),('thetaJN',float),('phiJL',float),('phi12',float),('psiJ',float)])
        for indx in np.arange(len(samples['psi'])):
            theta_jn, phi_JL, theta_1, theta_2, phi_12, a_1, a_2 = lalsim.SimInspiralTransformPrecessingWvf2PE( samples['incl'][indx], samples['a1x'][indx], samples['a1y'][indx], samples['a1z'][indx], samples['a2x'][indx], samples['a2y'][indx], samples['a2z'][indx], samples['m1'][indx], samples['m2'][indx],20, samples['phiorb'][indx])
            samples['thetaJN'][indx] = theta_jn
            samples['phiJL'][indx] = np.mod(phi_JL, 2*np.pi)
            samples['theta1'][indx]= theta_1
            samples['theta2'][indx] = theta_2
            samples['chi1'][indx] = a_1
            samples['chi2'][indx] = a_2
            samples['phi12'][indx] = np.mod(phi_12, 2*np.pi)


    # drop nonessential fields
    
    # Convert to dataframe
    samples_df = pd.DataFrame(samples)
    bilby_result = bilby.core.result.Result(label=fname,posterior=samples_df)
    return bilby_result

def load_priors(basedir):
    priors = bilby.gw.prior.BBHPriorDict(f"/home/colm.talbot/O4/pe_review/prior-files/{basedir}.prior")

    del priors["mass_1"], priors["mass_2"]
    for key in list(priors.keys()):
        if "time" in key:
            del priors[key]
        elif isinstance(priors[key], (float, int, bilby.core.prior.DeltaFunction)):
            del priors[key]

    if "a_1" in priors:
        priors["a_1"].latex_label = "$a_{1}$"
        priors["a_2"].latex_label = "$a_{2}$"
    if "lambda_1" in list(priors.keys()):
        priors["lambda_1"].latex_label = "$\\lambda_{1}$"
    if "lambda_2" in priors:
        priors["lambda_2"].latex_label = "$\\lambda_{2}$"
    return priors

def make_pp_plot(results, filename=None, save=True, confidence_interval=[0.68, 0.95, 0.997],
                 lines=None, legend_fontsize='x-small', keys=None, title=True,
                 confidence_interval_alpha=0.1, weight_list=None,
                 **kwargs):
    """
    Make a P-P plot for a set of runs with injected signals.

    Parameters
    ==========
    results: list
        A list of Result objects, each of these should have injected_parameters
    filename: str, optional
        The name of the file to save, the default is "outdir/pp.png"
    save: bool, optional
        Whether to save the file, default=True
    confidence_interval: (float, list), optional
        The confidence interval to be plotted, defaulting to 1-2-3 sigma
    lines: list
        If given, a list of matplotlib line formats to use, must be greater
        than the number of parameters.
    legend_fontsize: float
        The font size for the legend
    keys: list
        A list of keys to use, if None defaults to search_parameter_keys
    title: bool
        Whether to add the number of results and total p-value as a plot title
    confidence_interval_alpha: float, list, optional
        The transparency for the background condifence interval
    weight_list: list, optional
        List of the weight arrays for each set of posterior samples.
    kwargs:
        Additional kwargs to pass to matplotlib.pyplot.plot

    Returns
    =======
    fig, pvals:
        matplotlib figure and a NamedTuple with attributes `combined_pvalue`,
        `pvalues`, and `names`.
    """
    import matplotlib.pyplot as plt

    if keys is None:
        keys = results[0].search_parameter_keys

    if weight_list is None:
        weight_list = [None] * len(results)

    credible_levels = list()
    for i, result in enumerate(results):
        credible_levels.append(
            result.get_all_injection_credible_levels(keys, weights=weight_list[i])
        )
    credible_levels = pd.DataFrame(credible_levels)

    if lines is None:
        colors = ["C{}".format(i) for i in range(8)]
        linestyles = ["-", "--", ":"]
        lines = ["{}{}".format(a, b) for a, b in product(linestyles, colors)]
    if len(lines) < len(credible_levels.keys()):
        raise ValueError("Larger number of parameters than unique linestyles")

    x_values = np.linspace(0, 1, 1001)

    N = len(credible_levels)
    fig, ax = plt.subplots()

    if isinstance(confidence_interval, float):
        confidence_interval = [confidence_interval]
    if isinstance(confidence_interval_alpha, float):
        confidence_interval_alpha = [confidence_interval_alpha] * len(confidence_interval)
    elif len(confidence_interval_alpha) != len(confidence_interval):
        raise ValueError(
            "confidence_interval_alpha must have the same length as confidence_interval")

    for ci, alpha in zip(confidence_interval, confidence_interval_alpha):
        edge_of_bound = (1. - ci) / 2.
        lower = scipy.stats.binom.ppf(1 - edge_of_bound, N, x_values) / N
        upper = scipy.stats.binom.ppf(edge_of_bound, N, x_values) / N
        # The binomial point percent function doesn't always return 0 @ 0,
        # so set those bounds explicitly to be sure
        lower[0] = 0
        upper[0] = 0
        ax.fill_between(x_values, lower, upper, alpha=alpha, color='k')

    pvalues = []
    logger.info("Key: KS-test p-value")
    for ii, key in enumerate(credible_levels):
        pp = np.array([sum(credible_levels[key].values < xx) /
                       len(credible_levels) for xx in x_values])
        pvalue = scipy.stats.kstest(credible_levels[key], 'uniform').pvalue
        pvalues.append(pvalue)
        logger.info("{}: {}".format(key, pvalue))

        #try:
        #    name = results[0].priors[key].latex_label
        #except AttributeError:
        name = key
        label = "{} ({:2.3f})".format(name, pvalue)
        plt.plot(x_values, pp, lines[ii], label=label, **kwargs)

    Pvals = namedtuple('pvals', ['combined_pvalue', 'pvalues', 'names'])
    pvals = Pvals(combined_pvalue=scipy.stats.combine_pvalues(pvalues)[1],
                  pvalues=pvalues,
                  names=list(credible_levels.keys()))
    logger.info(
        "Combined p-value: {}".format(pvals.combined_pvalue))

    if title:
        ax.set_title("N={}, p-value={:2.4f}".format(
            len(results), pvals.combined_pvalue))
    ax.set_xlabel("C.I.")
    ax.set_ylabel("Fraction of events in C.I.")
    ax.legend(handlelength=2, labelspacing=0.25, fontsize=legend_fontsize)
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    fig.tight_layout()
    if save:
        if filename is None:
            filename = 'outdir/pp.png'
        safe_save_figure(fig=fig, filename=filename, dpi=500)

    return fig, pvals


parser = argparse.ArgumentParser()
parser.add_argument("--base-dir",default='my_set_16s',type=str)
parser.add_argument("--alt-coords",action='store_true')
opts = parser.parse_args()


# Can't use external code
#   - different param names
#   - coord system change too annoying
#base_name_short = opts.base_dir.split('_')[-1].replace('/','')
#priors = load_priors(base_name_short)
#print(priors)

if opts.alt_coords:
    param_names = param_names_alt

results = []
fname_list = glob.glob(opts.base_dir+'/analysis_*/rundir/extrinsic_posterior_samples.dat')
#print(fname_list)
indx_name =0
for fname in fname_list:
    indx_event = int(fname.split('/')[-3].replace('analysis_event_',''))
    print(fname,indx_event)
    # load in data
    result = load_result(fname,alt_coords=opts.alt_coords)
    #result.priors = priors
    
    # Add injection
    fname_inj = opts.base_dir+"/mdc.xml.gz"#fname.replace('rundir/extrinsic_posterior_samples.dat', 'mdc.xml.gz')
    #print(fname_inj)
    P = lalsimutils.xml_to_ChooseWaveformParams_array(fname_inj)[indx_event]
    P.fref = fref
    inj_vals = {}
    for name in param_names:
        scale = 1
        if name =='mc':
            scale = lal.MSUN_SI
        if name =='distance':
            scale =lal.PC_SI*1e6
        if name != 'phi12':
            inj_vals[name ] = P.extract_param(remap_export_to_P[name])/scale
        else:
            inj_vals[name] = P.extract_param('phi2') - P.extract_param('phi1')
    #inj_vals = [P.extract_param(remap_export_to_P[x]) for x in param_names]

    # add parameters to file
    result.injection_parameters = inj_vals

    #  augment list
    results.append(result)

#bilby.core.result.
make_pp_plot(results,"output_pp_bilby.pdf",keys=param_names)



"""test_like_singletemplate_emcee.py
   a test to see how to glue emcee to our likelihood evaluation.
"""

from __future__ import print_function

import sys
from optparse import OptionParser

import numpy
from matplotlib import pylab as plt

from glue.lal import Cache
import lalsimutils

from factored_likelihood import *
from ourio import *
import emcee

checkInputs = False

rosUseDifferentWaveformLengths = False    
rosUseRandomTemplateStartingFrequency = False

rosUseTargetedDistance = True
rosUseStrongPriorOnParameters = True
rosDebugMessages = True
rosShowSamplerInputDistributions = True
rosShowRunningConvergencePlots = True
rosShowTerminalSampleHistograms = True
rosSaveHighLikelihoodPoints = True

theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 
tEventFiducial = 0                                                                 # relative to GPS reference

optp = OptionParser()
optp.add_option("-c", "--cache-file", default=None, help="LIGO cache file containing all data needed.")
optp.add_option("-C", "--channel-name", action="append", help="instrument=channel-name, e.g. H1=FAKE-STRAIN. Can be given multiple times for different instruments.")
opts, args = optp.parse_args()

det_dict = {}
rhoExpected ={}
if opts.channel_name is not None and opts.cache_file is None:
    print("Cache file required when requesting channel data.", file=sys.stderr)
    exit(-1)
elif opts.channel_name is not None:
    det_dict = dict(map(lambda cname: cname.split("="), opts.channel_name))

Niter = 5 # Number of times to call likelihood function
Tmax = 0.05 # max ref. time
Tmin = -0.05 # min ref. time
Dmax = 110. * 1.e6 * lal.LAL_PC_SI # max distance
Dmin = 1. * 1.e6 * lal.LAL_PC_SI   # min distance

ampO =0 # sets which modes to include in the physical signal
Lmax = 2 # sets which modes to include
fref = 100
fminWavesSignal = 25  # too long can be a memory and time hog, particularly at 16 kHz
fSample = 4096*4

if rosUseDifferentWaveformLengths: 
    fminWavesTemplate = fminWavesSignal+0.005
else:
    if rosUseRandomTemplateStartingFrequency:
         print("   --- Generating a random template starting frequency  ---- ")
         fminWavesTemplate = fminWavesSignal+5.*np.random.random_sample()
    else:
        fminWavesTemplate = fminWavesSignal

#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
if len(det_dict) > 0:
    for d, chan in det_dict.items():
        data_dict[d] = lalsimutils.frame_data_to_hoff(opts.cache_file, chan)
else:
    m1 = 4*lal.LAL_MSUN_SI
    m2 = 3*lal.LAL_MSUN_SI

    Psig = ChooseWaveformParams(
        m1 = m1,m2 =m2,
        fmin = fminWavesSignal, 
        fref=fref, ampO=ampO,
                                radec=True, theta=1.2, phi=2.4,
                                detector='H1', 
                                dist=25.*1.e6*lal.LAL_PC_SI,
                                tref = theEpochFiducial,
        deltaT=1./fSample
                                )
    df = findDeltaF(Psig)
    Psig.print_params()
    Psig.deltaF = df
    data_dict['H1'] = lalsimutils.non_herm_hoff(Psig)
    Psig.detector = 'L1'
    data_dict['L1'] = lalsimutils.non_herm_hoff(Psig)
    Psig.detector = 'V1'
    data_dict['V1'] = lalsimutils.non_herm_hoff(Psig)

# TODO: Read PSD from XML
psd_dict = {}
analyticPSD_Q = True # For simplicity, using an analytic PSD
psd_dict['H1'] = lal.LIGOIPsd
psd_dict['L1'] = lal.LIGOIPsd
psd_dict['V1'] = lal.LIGOIPsd

print(" == Data report == ")
detectors = data_dict.keys()
rho2Net = 0
print(" Amplitude report :")
fminSNR =30
for det in detectors:
    IP = ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=Psig.deltaF,psd=psd_dict[det])
    rhoExpected[det] = rhoDet = IP.norm(data_dict[det])
    rho2Net += rhoDet*rhoDet
    print(det, " rho = ", rhoDet)
print("Network : ", np.sqrt(rho2Net))

if checkInputs:
    print(" == Plotting detector data (time domain; requires regeneration, MANUAL TIMESHIFTS,  and seperate code path! Argh!) == ")
    P = Psig.copy()
    P.tref = Psig.tref
    for det in detectors:
        P.detector=det   # we do 
        hT = hoft(P)
        tvals = float(P.tref - theEpochFiducial) + hT.deltaT*np.arange(len(hT.data.data))
        plt.figure(1)
        plt.plot(tvals, hT.data.data,label=det)

    tlen = hT.deltaT*len(np.nonzero(np.abs(hT.data.data)))
    tRef = np.abs(float(hT.epoch))
    plt.xlim( tRef-0.5,tRef+0.1)  # Not well centered, based on epoch to identify physical merger time
    plt.savefig("test_like_and_samp-input-hoft.pdf")



# Struct to hold template parameters
P = ChooseWaveformParams(fmin=fminWavesTemplate, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         fref=fref,
         tref=theEpochFiducial,
         deltaT=1./fSample,
         dist=100*1.e6*lal.LAL_PC_SI,
         deltaF=df)


#
# Perform the Precompute stage
#
rholms_intp, crossTerms, rholms, epoch_post = PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict,psd_dict, Lmax, analyticPSD_Q)
print("Finished Precomputation...")
print("====Generating metadata from precomputed results =====")
distBoundGuess = estimateUpperDistanceBoundInMpc(rholms, crossTerms)
print(" distance probably less than ", distBoundGuess, " Mpc")

nEvals = 0
def likelihood_function( x):
    global nEvals
    global rholms_intp
    global crossTerms
    global Lmax
    phi, theta, tref, phiref, incl, psi, dist = x

    # Eliminate  implausible values
    # These don't count as evaluations
    if np.abs(tref)>tWindowExplore[1] or dist < 0 or dist > distBoundGuess or psi<0 or psi>np.pi or phiref<0 or phiref>2*np.pi or phi<0 or phi>2*np.pi:
        return -np.inf

    P.phi = phi # right ascension
    P.theta = theta # declination
    P.tref = theEpochFiducial + tref # ref. time (rel to epoch for data taking)
    P.phiref = phiref # ref. orbital phase
    P.incl = incl # inclination
    P.psi = psi # polarization angle
    P.dist = dist*(1e6*lal.LAL_PC_SI) # luminosity distance

    lnL = FactoredLogLikelihood(P, rholms_intp, crossTerms, Lmax)
    if rosDebugMessages:
        if (numpy.mod(nEvals,400)==10 and nEvals>100):
            print("\t Params ", nEvals, " (RA, DEC, tref, phiref, incl, psi, dist) =")
            print("\t", nEvals, P.phi, P.theta, float(P.tref-theEpochFiducial), P.phiref, P.incl, P.psi, P.dist/(1e6*lal.LAL_PC_SI), lnL)

    nEvals+=1
    return np.exp(lnL)




## ALTERNATIVE: Just in case we are not being very efficient,let's compare an MCMC.
# emcee
ndim = 7
nwalkers = 100
p0 = [[np.random.random_sample(),np.random.random_sample(),0.002*np.random.random_sample(),np.random.random_sample(),np.random.random_sample(),np.random.random_sample(),np.random.random_sample()*distBoundGuess] for i in range(nwalkers)]

sampler = emcee.EnsembleSampler(nwalkers, ndim, likelihood_function,threads=2)
sampler.run_mcmc(p0,2000)

np.savetxt('test-emcee.dat', sampler.flatchain)




#
# kde.py: KDE utilities for density estimation in unusual topologies.
#
# Copyright 2012 Will M. Farr <will.farr@ligo.org>
# Modified 2017 Leo P. Singer <leo.singer@ligo.org> to handle 1D KDEs
# gracefully.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from __future__ import division
import numpy as np
from scipy.stats import gaussian_kde


class BoundedKDE(gaussian_kde):
    """Density estimation using a KDE on bounded domains.
    Bounds can be any combination of low or high (if no bound, set to
    ``float('inf')`` or ``float('-inf')``), and can be periodic or
    non-periodic.  Cannot handle topologies that have
    multi-dimensional periodicities; will only handle topologies that
    are direct products of (arbitrary numbers of) R, [0,1], and S1.
    :param pts:
        ``(Ndim, Npts)`` shaped array of points (as in :class:`gaussian_kde`).
    :param low: 
        Lower bounds; if ``None``, assume no lower bounds.
    :param high:
        Upper bounds; if ``None``, assume no upper bounds.
    :param periodic:
        Boolean array giving periodicity in each dimension; if
        ``None`` assume no dimension is periodic.
    :param bw_method: (optional)
        Bandwidth estimation method (see :class:`gaussian_kde`)."""
    def __init__(self, pts, low=-np.inf, high=np.inf, periodic=False,
                 bw_method=None):

        super(BoundedKDE, self).__init__(pts, bw_method=bw_method)
        self._low = np.broadcast_to(
            low, self.d).astype(self.dataset.dtype)
        self._high = np.broadcast_to(
            high, self.d).astype(self.dataset.dtype)
        self._periodic = np.broadcast_to(
            periodic, self.d).astype(bool)

    def evaluate(self, pts):
        """Evaluate the KDE at the given points."""

        pts = np.atleast_2d(pts)
        d, m = pts.shape
        if d != self.d and d == 1 and m == self.d:
            pts = pts.T

        pts_orig = pts
        pts = np.copy(pts_orig)

        den = super(BoundedKDE, self).evaluate(pts)

        for i, (low, high, period) in enumerate(zip(self._low, self._high,
                                                    self._periodic)):
            if period:
                P = high - low
                
                pts[i, :] += P
                den += super(BoundedKDE, self).evaluate(pts)

                pts[i,:] -= 2.0*P
                den += super(BoundedKDE, self).evaluate(pts)

                pts[i,:] = pts_orig[i,:]

            else:
                if not np.isneginf(low):
                    pts[i,:] = 2.0*low - pts[i,:]
                    den += super(BoundedKDE, self).evaluate(pts)
                    pts[i,:] = pts_orig[i,:]

                if not np.isposinf(high):
                    pts[i,:] = 2.0*high - pts[i,:]
                    den += super(BoundedKDE, self).evaluate(pts)
                    pts[i,:] = pts_orig[i,:]

        return den

    __call__ = evaluate

    def quantile(self, pt):
        """Quantile of ``pt``, evaluated by a greedy algorithm.
        :param pt:
            The point at which the quantile value is to be computed.
        The quantile of ``pt`` is the fraction of points used to
        construct the KDE that have a lower KDE density than ``pt``."""

        return np.count_nonzero(self(self.dataset) < self(pt)) / self.n



# Special case
class BoundedKDE_2d(gaussian_kde):
    r"""Represents a two-dimensional Gaussian kernel density estimator
    for a probability distribution function that exists on a bounded
    domain."""

    def __init__(self, pts, xlow=None, xhigh=None, ylow=None, yhigh=None, *args, **kwargs):
        """Initialize with the given bounds.  Either ``low`` or
        ``high`` may be ``None`` if the bounds are one-sided.  Extra
        parameters are passed to :class:`gaussian_kde`.

        :param xlow: The lower x domain boundary.

        :param xhigh: The upper x domain boundary.

        :param ylow: The lower y domain boundary.

        :param yhigh: The upper y domain boundary.
        """
        pts = np.atleast_2d(pts)

        assert pts.ndim == 2, 'Bounded_kde can only be two-dimensional'

        super(BoundedKDE_2d, self).__init__(pts.T, *args, **kwargs)

        self._xlow = xlow
        self._xhigh = xhigh
        self._ylow = ylow
        self._yhigh = yhigh

    @property
    def xlow(self):
        """The lower bound of the x domain."""
        return self._xlow

    @property
    def xhigh(self):
        """The upper bound of the x domain."""
        return self._xhigh

    @property
    def ylow(self):
        """The lower bound of the y domain."""
        return self._ylow

    @property
    def yhigh(self):
        """The upper bound of the y domain."""
        return self._yhigh

    def evaluate(self, pts):
        """Return an estimate of the density evaluated at the given
        points."""
        pts = np.atleast_2d(pts)
        assert pts.ndim == 2, 'points must be two-dimensional'

        x, y = pts.T
        pdf = super(BoundedKDE_2d, self).evaluate(pts.T)
        if self.xlow is not None:
            pdf += super(BoundedKDE_2d, self).evaluate([2*self.xlow - x, y])

        if self.xhigh is not None:
            pdf += super(BoundedKDE_2d, self).evaluate([2*self.xhigh - x, y])

        if self.ylow is not None:
            pdf += super(BoundedKDE_2d, self).evaluate([x, 2*self.ylow - y])

        if self.yhigh is not None:
            pdf += super(BoundedKDE_2d, self).evaluate([x, 2*self.yhigh - y])

        if self.xlow is not None:
            if self.ylow is not None:
                pdf += super(BoundedKDE_2d, self).evaluate([2*self.xlow - x, 2*self.ylow - y])

            if self.yhigh is not None:
                pdf += super(BoundedKDE_2d, self).evaluate([2*self.xlow - x, 2*self.yhigh - y])

        if self.xhigh is not None:
            if self.ylow is not None:
                pdf += super(BoundedKDE_2d, self).evaluate([2*self.xhigh - x, 2*self.ylow - y])
            if self.yhigh is not None:
                pdf += super(BoundedKDE_2d, self).evaluate([2*self.xhigh - x, 2*self.yhigh - y])

        return pdf

    def __call__(self, pts):
        pts = np.atleast_2d(pts)
        out_of_bounds = np.zeros(pts.shape[0], dtype='bool')

        if self.xlow is not None:
            out_of_bounds[pts[:, 0] < self.xlow] = True
        if self.xhigh is not None:
            out_of_bounds[pts[:, 0] > self.xhigh] = True
        if self.ylow is not None:
            out_of_bounds[pts[:, 1] < self.ylow] = True
        if self.yhigh is not None:
            out_of_bounds[pts[:, 1] > self.yhigh] = True

        results = self.evaluate(pts)
        results[out_of_bounds] = 0.
        return results




import numpy as np
import RIFT.lalsimutils as lalsimutils
remap_ILE_2_LI = {
 "s1z":"a1z", "s2z":"a2z", 
 "s1x":"a1x", "s1y":"a1y",
 "s2x":"a2x", "s2y":"a2y",
 "chi1_perp":"chi1_perp",
 "chi2_perp":"chi2_perp",
 "chi1":'a1',
 "chi2":'a2',
 "cos_phiJL": 'cos_phiJL',
 "sin_phiJL": 'sin_phiJL',
 "cos_theta1":'costilt1',
 "cos_theta2":'costilt2',
 "theta1":"tilt1",
 "theta2":"tilt2",
  "xi":"chi_eff", 
  "chiMinus":"chi_minus", 
  "delta":"delta", 
  "delta_mc":"delta", 
 "mtot":'mtotal', "mc":"mc", "eta":"eta","m1":"m1","m2":"m2",
  "cos_beta":"cosbeta",
  "beta":"beta",
  "LambdaTilde":"lambdat",
  "DeltaLambdaTilde": "dlambdat",
  "thetaJN":"theta_jn"}
remap_LI_to_ILE = { "a1z":"s1z", "a2z":"s2z", "chi_eff":"xi", "lambdat":"LambdaTilde", 'mtotal':'mtot', "distance":"dist", 'ra':'phi', 'dec':'theta',"phiorb":"phiref"}

import numpy.lib.recfunctions as rfn

def extract_combination_from_LI(samples_LI, p):
    """
    extract_combination_from_LI
      - reads in known columns from posterior samples
      - for selected known combinations not always available, it will compute them from standard quantities
    """
    if p in samples_LI.dtype.names:  # e.g., we have precomputed it
        return samples_LI[p]
    if p in remap_ILE_2_LI.keys():
       if remap_ILE_2_LI[p] in samples_LI.dtype.names:
         return samples_LI[ remap_ILE_2_LI[p] ]
    if (p == 'chi_eff' or p=='xi') and 'a1z' in samples_LI.dtype.names:
         m1 = samples_LI['m1']
         m2 = samples_LI['m2']
         a1z = samples_LI['a1z']
         a2z = samples_LI['a2z']
         return (m1 * a1z + m2*a2z)/(m1+m2)
    # Return cartesian components of spin1, spin2.  NOTE: I may already populate these quantities in 'Add important quantities'
    if p == 'chiz_plus':
        print(" Transforming ")
        if 'a1z' in samples_LI.dtype.names:
            return (samples_LI['a1z']+ samples_LI['a2z'])/2.
        if 'theta1' in samples_LI.dtype.names:
            return (samples_LI['a1']*np.cos(samples_LI['theta1']) + samples_LI['a2']*np.cos(samples_LI['theta2']) )/2.
#        return (samples_LI['a1']+ samples_LI['a2'])/2.
    if p == 'chiz_minus':
        print(" Transforming ")
        if 'a1z' in samples_LI.dtype.names:
            return (samples_LI['a1z']- samples_LI['a2z'])/2.
        if 'theta1' in samples_LI.dtype.names:
            return (samples_LI['a1']*np.cos(samples_LI['theta1']) - samples_LI['a2']*np.cos(samples_LI['theta2']) )/2.
#        return (samples_LI['a1']- samples_LI['a2'])/2.
    if  'theta1' in samples_LI.dtype.names:
        if p == 's1x':
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) * np.cos( samples_LI['phi1'])
        if p == 's1y' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) * np.sin( samples_LI['phi1'])
        if p == 's2x':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) * np.cos( samples_LI['phi2'])
        if p == 's2y':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) * np.sin( samples_LI['phi2'])
        if p == 'chi1_perp' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) 
        if p == 'chi2_perp':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) 
    if 'lambdat' in samples_LI.dtype.names:  # LI does sampling in these tidal coordinates
        lambda1, lambda2 = lalsimutils.tidal_lambda_from_tilde(samples_LI["m1"], samples_LI["m2"], samples_LI["lambdat"], samples_LI["dlambdat"])
        if p == "lambda1":
            return lambda1
        if p == "lambda2":
            return lambda2
    if p == 'delta' or p=='delta_mc':
        return (samples_LI['m1']  - samples_LI['m2'])/((samples_LI['m1']  + samples_LI['m2']))
    # Return cartesian components of Lhat
    if p == 'product(sin_beta,sin_phiJL)':
        return np.sin(samples_LI[ remap_ILE_2_LI['beta'] ]) * np.sin(  samples_LI['phi_jl'])
    if p == 'product(sin_beta,cos_phiJL)':
        return np.sin(samples_LI[ remap_ILE_2_LI['beta'] ]) * np.cos(  samples_LI['phi_jl'])

    if p == 'mc':
        m1v= samples_LI["m1"]
        m2v = samples_LI["m2"]
        return lalsimutils.mchirp(m1v,m2v)
    if p == 'eta':
        m1v= samples_LI["m1"]
        m2v = samples_LI["m2"]
        return lalsimutils.symRatio(m1v,m2v)

    if p == 'phi1':
        return np.angle(samples_LI['a1x']+1j*samples_LI['a1y'])
    if p == 'chi_pavg':
        samples = np.array([samples_LI["m1"], samples_LI["m2"], samples_LI["a1x"], samples_LI["a1y"], samples_LI["a1z"], samples_LI["a2x"], samples_LI["a2y"], samples_LI["a2z"]]).T
        with Pool(12) as pool:   
            chipavg = np.array(pool.map(fchipavg, samples))          
        return chipavg

    if p == 'chi_p':
        samples = np.array([samples_LI["m1"], samples_LI["m2"], samples_LI["a1x"], samples_LI["a1y"], samples_LI["a1z"], samples_LI["a2x"], samples_LI["a2y"], samples_LI["a2z"]]).T
        with Pool(12) as pool:   
            chip = np.array(pool.map(fchip, samples))          
        return chip

    # Backup : access lambdat if not present
    if (p == 'lambdat' or p=='dlambdat') and 'lambda1' in samples_LI.dtype.names:
        Lt,dLt = lalsimutils.tidal_lambda_tilde(samples_LI['m1'], samples_LI['m2'],  samples_LI['lambda1'], samples_LI['lambda2'])
        if p=='lambdat':
            return Lt
        if p=='dlambdat':
            return dLt

    if p == "q"  and 'm1' in samples_LI.dtype.names:
        return samples_LI["m2"]/samples_LI["m1"]

    if 'inverse(' in p:
        # Drop first and last characters
        a=p.replace(' ', '') # drop spaces
        a = a[:len(a)-1] # drop last
        a = a[8:]
        if a =='q' and 'm1' in samples_LI.dtype.names:
            return samples_LI["m1"]/samples_LI["m2"]

    print(" No access for parameter ", p)
    return np.zeros(len(samples_LI['m1']))  # to avoid causing a hard failure

def add_field(a, descr):
    """Return a new array that is like "a", but has additional fields.

    Arguments:
      a     -- a structured numpy array
      descr -- a numpy type description of the new fields

    The contents of "a" are copied over to the appropriate fields in
    the new array, whereas the new fields are uninitialized.  The
    arguments are not modified.

    >>> sa = numpy.array([(1, 'Foo'), (2, 'Bar')], \
                         dtype=[('id', int), ('name', 'S3')])
    >>> sa.dtype.descr == numpy.dtype([('id', int), ('name', 'S3')])
    True
    >>> sb = add_field(sa, [('score', float)])
    >>> sb.dtype.descr == numpy.dtype([('id', int), ('name', 'S3'), \
                                       ('score', float)])
    True
    >>> numpy.all(sa['id'] == sb['id'])
    True
    >>> numpy.all(sa['name'] == sb['name'])
    True
    """
    if a.dtype.fields is None:
        raise ValueError("`A' must be a structured numpy array")
    b = np.empty(a.shape, dtype=a.dtype.descr + descr)
    for name in a.dtype.names:
        b[name] = a[name]
    return b





def standard_expand_samples(samples):
    """
    Do some things which add a bunch of standard fields to the samples, if I don't have them.  
    Used in plot_posterior_corner.py for example
    """
    if not 'mtotal' in samples.dtype.names and 'mc' in samples.dtype.names:  # raw LI samples use 
        q_here = samples['q']
        eta_here = q_here/(1+q_here)
        mc_here = samples['mc']
        mtot_here = mc_here / np.power(eta_here, 3./5.)
        m1_here = mtot_here/(1+q_here)
        samples = add_field(samples, [('mtotal', float)]); samples['mtotal'] = mtot_here
        samples = add_field(samples, [('eta', float)]); samples['eta'] = eta_here
        samples = add_field(samples, [('m1', float)]); samples['m1'] = m1_here
        samples = add_field(samples, [('m2', float)]); samples['m2'] = mtot_here * q_here/(1+q_here)
        
    if "theta1" in samples.dtype.names and not('chi1_perp' in samples.dtype.names):
        a1x_dat = samples["a1"]*np.sin(samples["theta1"])*np.cos(samples["phi1"])
        a1y_dat = samples["a1"]*np.sin(samples["theta1"])*np.sin(samples["phi1"])
        chi1_perp = samples["a1"]*np.sin(samples["theta1"])

        a2x_dat = samples["a2"]*np.sin(samples["theta2"])*np.cos(samples["phi2"])
        a2y_dat = samples["a2"]*np.sin(samples["theta2"])*np.sin(samples["phi2"])
        chi2_perp = samples["a2"]*np.sin(samples["theta2"])

                                      
        samples = add_field(samples, [('a1x', float)]);  samples['a1x'] = a1x_dat
        samples = add_field(samples, [('a1y', float)]); samples['a1y'] = a1y_dat
        samples = add_field(samples, [('a2x', float)]);  samples['a2x'] = a2x_dat
        samples = add_field(samples, [('a2y', float)]);  samples['a2y'] = a2y_dat
        samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
        samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp
        if not 'chi_eff' in samples.dtype.names:
            samples = add_field(samples, [('chi_eff',float)]); samples['chi_eff'] = (samples["m1"]*samples["a1z"]+samples["m2"]*samples["a2z"])/(samples["m1"]+samples["m2"])
 
    elif 'a1x' in samples.dtype.names and not 'chi1_perp' in samples.dtype.names:
        chi1_perp = np.sqrt(samples['a1x']**2 + samples['a1y']**2)
        chi2_perp = np.sqrt(samples['a2x']**2 + samples['a2y']**2)
        samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
        samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp

        # Askold: this part will check if phi1, phi2, phi12 are present. If not, compute and add the missing ones
        phi_fields = ['phi1', 'phi2', 'phi12']
        phi_func_dict = {
            'phi1': lambda samples: np.arctan2(samples['a1x'], samples['a1y']),
            'phi2': lambda samples: np.arctan2(samples['a2x'], samples['a2y']),
            'phi12': lambda samples: samples['phi2'] - samples['phi1']
        }

        for field_name in phi_fields:
            if not (field_name in samples.dtype.names):
                samples = add_field(samples, [(field_name, float)])
                samples[field_name] = phi_func_dict[field_name](samples)

    if 'lambda1' in samples.dtype.names and not ('lambdat' in samples.dtype.names):
        Lt,dLt = lalsimutils.tidal_lambda_tilde(samples['m1'], samples['m2'],  samples['lambda1'], samples['lambda2'])
        samples = add_field(samples, [('lambdat', float)]); samples['lambdat'] = Lt
        samples = add_field(samples, [('dlambdat', float)]); samples['dlambdat'] = dLt


    return samples


######### MUlTIPROCESSING FUNCTIONS ############
from multiprocessing import Pool 

def fchipavg(sample):
            P=lalsimutils.ChooseWaveformParams()
            P.m1 = sample[0]
            P.m2 = sample[1]
            P.s1x = sample[2]
            P.s1y = sample[3]
            P.s1z = sample[4]
            P.s2x = sample[5]
            P.s2y = sample[6]
            P.s2z = sample[7]
            if (P.s1x == 0 and P.s1y == 0 and P.s2x == 0 and P.s2y == 0):
                chipavg = 0
            elif (P.s1x == 0 and P.s1y == 0 and P.s1z == 0) or (P.s2x == 0 and P.s2y == 0 and P.s2z == 0):
                chipavg = P.extract_param('chi_p')
            else:
                chipavg = P.extract_param('chi_pavg')
            return chipavg     

def fchip(sample):
            P=lalsimutils.ChooseWaveformParams()
            P.m1 = sample[0]
            P.m2 = sample[1]
            P.s1x = sample[2]
            P.s1y = sample[3]
            P.s1z = sample[4]
            P.s2x = sample[5]
            P.s2y = sample[6]
            P.s2z = sample[7]
            chip = P.extract_param('chi_p')
            return chip  

def dump_pesummary_samples_to_file_as_rift(fname_h5,key,fname_out):
    """
    >>> import samples_utils
    >>> samples_utils.dump_pesummary_samples_to_file_as_rift("metafile.h5", "bilby-IMRPhenomXPHM-SpinTaylor-3",'test.dat')
    $ convert_output_format_inference2ile --posterior-samples test.dat --output-xml my.xml.gz
    """
    import h5py
    BBH = h5py.File(fname_h5, 'r')
    if not (key in BBH.keys()):
        raise Exception(" Unknown key in file ", key, fname_h5)
    post_key = 'posterior'
    if 'posterior_samples' in BBH[key]:
        post_key = 'posterior_samples'
    samples = BBH[key][post_key]
    if hasattr(BBH[key][post_key], 'dtype'):
        dtype_us = BBH[key][post_key].dtype #
    else:
        dtype_us = np.dtype(list(map( lambda x: (x,float), list(samples.keys() ))))  # old style
    npts = len(BBH[key][post_key]['mass_1'])
    # cast to conventional structure so we can call recfunctions
    samp = np.zeros(npts,dtype=dtype_us)
    for name in dtype_us.names:
        if 'snr' in name:
            continue
        if 'approximant' in name:
            continue
        samp[name] = samples[name]
    # Rename
    samp = rfn.rename_fields(samp, {'chirp_mass':'mc', 'mass_1':'m1', 'mass_2':'m2','geocent_time':'time','luminosity_distance':'distance','phase':'phiorb','chi_1_in_plane':'chi1_perp','spin_1x': 'a1x', 'spin_1y':'a1y', 'spin_2x': 'a2x','spin_2y':'a2y', 'spin_1z':'a1z', 'spin_2z':'a2z', 'chi_2_in_plane':'chi2_perp','iota':'incl','lambda_1':'lambda1', 'lambda_2':'lambda2','lambdat':'LambdaTilde'})
    # Drop
    ugly_fields = [x for x in samp.dtype.names if 'recalib' in x or 'snr' in x]
    samp = rfn.drop_fields(samp,ugly_fields)

    np.savetxt(fname_out,samp,header=" ".join(samp.dtype.names) )




"""
test_profile_components.py: 
   - simple timing tests on different elements in the likelihood
"""
from pylal import Fr

import factored_likelihood 
import numpy as np
import lal
import lalsimulation as lalsim

import lalsimutils
#from factored_likelihood import * # gets most things
from matplotlib import pylab as plt
import sys
import ourparams
opts,  rosDebugMessagesDictionary = ourparams.ParseStandardArguments()
factored_likelihood.rosDebugMessagesDictionary = rosDebugMessagesDictionary
lalsimutils.rosDebugMessagesDictionary            = rosDebugMessagesDictionary

#
# Set by user
#
checkResults = True
rosUseZeroNoiseCache  = True
approxTemplate = lalsim.TaylorT4

cost_dict = {}
#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
data_dict_time = {}
psd_dict = {}
rhoExpected ={}
rhoManual ={}
rhoExpectedAlt ={}
analyticPSD_Q = True # For simplicity, using an analytic PSD

# Note: Injected signal has fmin = 25. We should choose a different value.
fminWaves = fminWavesSignal = fminWavesTemplate = opts.fmin_Template  # too long can be a memory and time hog, particularly at 16 kHz
fminSNR =opts.fmin_SNR
fSample = 16384 #opts.srate


theEpochFiducial = lal.LIGOTimeGPS(1000000014.000000000)   # Use actual injection GPS time (assumed from trigger)
#theEpochFiducial = lal.LIGOTimeGPS(1000000000.000000000)     # Use epoch of the data
tEventFiducial = 0   #  time relative to fiducial epoch, used to identify window to look in.  Checked empirically.

detectors = ['H1', "L1", "V1"]
psd_dict['H1'] = lalsim.SimNoisePSDiLIGOSRD
#psd_dict['L1'] = lalsim.SimNoisePSDiLIGOSRD
psd_dict['V1'] = lalsim.SimNoisePSDiLIGOSRD

Psig = lalsimutils.xml_to_ChooseWaveformParams_array("mdc.xml.gz")[0]  # Load in the physical parameters of the injection (=first element)
Psig.approx=lalsim.SEOBNRv4
Psig.print_params()


# Load IFO FFTs.
tS = lal.GPSTimeNow()
# ASSUME data has same sampling rate!
df=1./32
dt = 1./fSample
Psig.deltaF = df
Psig.detector='H1'
data_dict['H1'] = lalsimutils.non_herm_hoff(Psig)
#Psig.detector='L1'
#data_dict['L1'] = lalsimutils.non_herm_hoff(Psig)
Psig.detector='V1'
data_dict['V1'] = lalsimutils.non_herm_hoff(Psig)


tE = lal.GPSTimeNow()
cost_dict['readin'] = tE-tS
#print data_dict['H1'].data.data[10]  # confirm data loaded
df = data_dict['H1'].deltaF  
fSample = len(data_dict['H1'].data.data)*data_dict['H1'].deltaF  # Note two-sided
print(" sampling rate of data = ", fSample)




print(" ======= Template specified: precomputing all quantities ==========")
# Struct to hold template parameters
# Fiducial distance provided but will not be used
m1 = 4*lal.MSUN_SI
m2 = 4*lal.MSUN_SI
ampO =opts.amporder # sets which modes to include in the physical signal
Lmax = opts.Lmax # sets which modes to include
fref = opts.fref
P =  lalsimutils.ChooseWaveformParams(fmin=fminWaves, radec=False,
         m1=m1,m2=m2,
         ampO=ampO,
         approx=approxTemplate,
         fref=fref,
         deltaT=1./fSample,
         tref=theEpochFiducial,
         dist=100*1.e6*lal.PC_SI,  # critical that this is fiducial (will be reset later)
         deltaF=df)
tStartPrecompute = lal.GPSTimeNow()
rholms_intp, crossTerms, crossTermsV, rholms, rest = factored_likelihood.PrecomputeLikelihoodTerms(theEpochFiducial,0.1, P, data_dict,psd_dict, Lmax, 2000, analyticPSD_Q,ignore_threshold=opts.opt_SkipModeThreshold,inv_spec_trunc_Q=opts.psd_TruncateInverse,T_spec=opts.psd_TruncateInverseTime,NR_group=opts.NR_template_group,NR_param=opts.NR_template_param,use_external_EOB=opts.use_external_EOB,ROM_group=opts.ROM_template_group,ROM_param=opts.ROM_template_group)

# Pack operation does it for each detector, so I need a loop
lookupNKdict = {}
lookupKNdict={}
lookupKNconjdict={}
ctUArrayDict = {}
ctVArrayDict={}
rholmArrayDict={}
rholms_intpArrayDict={}
for det in rholms_intp.keys():
    lookupNKdict[det],lookupKNdict[det], lookupKNconjdict[det], ctUArrayDict[det], ctVArrayDict[det], rholmArrayDict[det], rholms_intpArrayDict[det] = factored_likelihood.PackLikelihoodDataStructuresAsArrays( rholms[det].keys(), rholms_intp[det], rholms[det], crossTerms[det])
#rholms_intp, crossTerms, rholms, epoch_post, lookupNKdict,lookupKNdict, ctUArrayDict, ctVArrayDict, rholmArrayDict, rholms_intpArrayDict = factored_likelihood.PrecomputeLikelihoodTerms(theEpochFiducial,P, data_dict, psd_dict, Lmax, analyticPSD_Q,ignore_threshold=opts.opt_SkipModeThreshold)
tEndPrecompute = lal.GPSTimeNow()
cost_dict['precompute'] = float(tEndPrecompute-tStartPrecompute)

tS = lal.GPSTimeNow()
nEvals = 7000
for i in np.arange(nEvals):
#    P = Psig.copy()
#    P.tref +=0.001*np.random.random_sample()  # mimic setting parameters in the structure
    lnL = factored_likelihood.FactoredLogLikelihood( Psig, rholms,  rholms_intp, crossTerms,crossTermsV, 2)
tE = lal.GPSTimeNow()
cost_dict['lnL'] = float(tE-tS)/nEvals

tS = lal.GPSTimeNow()
nEvals = 7000
for i in np.arange(nEvals):
    lnL = factored_likelihood.SingleDetectorLogLikelihoodData(theEpochFiducial, rholms_intp,Psig.tref, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, 'H1')
tE = lal.GPSTimeNow()
cost_dict['lnLdata1'] = float(tE-tS)/nEvals



tS = lal.GPSTimeNow()
nEvals = 70000
for i in np.arange(nEvals):
    lnL = factored_likelihood.SingleDetectorLogLikelihoodDataViaArray(theEpochFiducial,lookupNKdict, rholms_intpArrayDict,Psig.tref, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist,  'H1')
tE = lal.GPSTimeNow()
cost_dict['lnLdata1b'] = float(tE-tS)/nEvals
print("  +++ Fast lnLData eval (array-ized)++")
print(" Per evaluation, not vectorizing over time, cost is ", cost_dict['lnLdata1b'])


tS = lal.GPSTimeNow()
nEvals = int(fSample*5)   # should be a controlled number of evals, in case data set gets large
for i in np.arange(nEvals):
    lnL = factored_likelihood.DiscreteSingleDetectorLogLikelihoodDataViaArray(theEpochFiducial,lookupNKdict, rholmArrayDict, 1./fSample, Psig.tref, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist,  'H1')
tE = lal.GPSTimeNow()
cost_dict['lnLdataDiscArrayPerTime'] = float(tE-tS)/nEvals / len(rholmArrayDict['H1'][0])  # Cost *per time bin* being evaluated
cost_dict['lnLdataDiscArray'] = float(tE-tS)/nEvals 
print("  +++ Fast lnLdata eval ++")
print("  In total ", cost_dict['lnLdataDiscArray'],  " per evaluation of the entire time array, using nEvals = ", nEvals)
print("  Per time point (npts =  ", len(rholmArrayDict['H1'][0]), " )  the cost is ", cost_dict['lnLdataDiscArrayPerTime'])



tS = lal.GPSTimeNow()
nEvals = 100000
for i in np.arange(nEvals):
    lnL = factored_likelihood.SingleDetectorLogLikelihoodModelViaArray(lookupNKdict, ctUArrayDict, ctVArrayDict,  Psig.tref, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist,  'H1')
tE = lal.GPSTimeNow()
cost_dict['lnLModelArray'] = float(tE-tS)/nEvals  # Cost *per time bin* being evaluated
print("  +++ Fast lnLModel eval ++")
print(" Per evaluation, cost is ", cost_dict['lnLModelArray'])


if False:
    tS = lal.GPSTimeNow()
    nEvals = 10
    npts = 10
    P = Psig.copy()
    phi = np.ones(npts)*P.phi
    theta = np.ones(npts)*P.theta
    tref = np.ones(npts)*(P.tref-theEpochFiducial)
    phiref = np.ones(npts)*P.phiref
    incl = np.ones(npts)*P.incl
    psi = np.ones(npts)*P.psi
    dist = np.ones(npts)*P.dist
    for i in np.arange(nEvals):
        lnL =factored_likelihood.FactoredLogLikelihoodVectorized(theEpochFiducial,phi,theta,tref, phiref, incl, psi, dist, rholms_intp, crossTerms,Psig.tref,  Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, detectors)
    tE = lal.GPSTimeNow()
    cost_dict['lnLVector'] = float(tE-tS)/nEvals/npts
    print("  +++ Fast lnLVector eval ++")
    print("  In total ", cost_dict['lnLVector'],  " per evaluation of each point, using nEvals*npts = ", nEvals*npts)



tS = lal.GPSTimeNow()
nEvals = 7000
for i in np.arange(nEvals):
    lnL = factored_likelihood.SingleDetectorLogLikelihoodModel(crossTerms,crossTermsV, Psig.tref, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, 'H1')
tE = lal.GPSTimeNow()
cost_dict['lnLmodel1'] = float(tE-tS)/nEvals

tS = lal.GPSTimeNow()
nEvals = 100000
for i in np.arange(nEvals):
        Ylms = factored_likelihood.ComputeYlms(Lmax, 0., 0.)
tE = lal.GPSTimeNow()
cost_dict['Ylms'] = float(tE-tS)/nEvals

tS = lal.GPSTimeNow()
nEvals = 100000
for i in np.arange(nEvals):
        rho22 = rholms_intp['H1'][(2,2)](0.)
tE = lal.GPSTimeNow()
cost_dict['rholms'] = float(tE-tS)/nEvals

tS = lal.GPSTimeNow()
nEvals = 10000
for i in np.arange(nEvals):
        np.roll(data_dict['H1'],500)
tE = lal.GPSTimeNow()
cost_dict['roll'] = float(tE-tS)/nEvals

if False:
    tS = lal.GPSTimeNow()
    nEvals = 1000
    for i in np.arange(nEvals):
        lnL =factored_likelihood.NetworkLogLikelihoodPolarizationMarginalized(theEpochFiducial, rholms_intp, crossTerms,Psig.tref,  Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, detectors)
    tE = lal.GPSTimeNow()
    cost_dict['psi'] = float(tE-tS)/nEvals



# tS = lal.GPSTimeNow()
# nEvals = 20
# for i in np.arange(nEvals):
#     lnLmargTimeD = factored_likelihood.NetworkLogLikelihoodTimeMarginalizedDiscrete(theEpochFiducial, rholms, crossTerms,Psig.tref,factored_likelihood.tWindowExplore, Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, detectors)
# tE = lal.GPSTimeNow()
# cost_dict['lnLmargTimeDiscrete'] = float(tE-tS)/nEvals

if False:
    tS = lal.GPSTimeNow()
    nEvals = 20
    for i in np.arange(nEvals):
        lnLmargTime = factored_likelihood.NetworkLogLikelihoodTimeMarginalized(theEpochFiducial, rholms_intp, crossTerms,Psig.tref, factored_likelihood.tWindowExplore,Psig.phi,Psig.theta, P.incl, P.phiref,P.psi, P.dist, 2, detectors)
    tE = lal.GPSTimeNow()
    cost_dict['lnLmargTime'] = float(tE-tS)/nEvals


print(" ======= Report ==========")
print("Sampling rate: ", fSample , " and data duration ", len(data_dict['H1'].data.data)/fSample)
print("Reading data :", float(cost_dict['readin']))
print("Precomputation time (per intrinsic parameters!) : ", cost_dict['precompute'])
print("Precomputation ingredient: rolling the data : ", cost_dict['roll'])
print("Per-evaluation FactoredLogLikelihood : ", cost_dict['lnL'])
print("Per-evaluation SingleDetectorLogLikelihoodData  *3: ", cost_dict['lnLdata1']*3)
print("Per-evaluation DiscreteSingleDetectorLogLikelihoodDataViaArray  *3: ", cost_dict['lnLdataDiscArrayPerTime']*3)
print("Per-evaluation SingleDetectorLogLikelihoodModel *3 : ", cost_dict['lnLmodel1']*3)
print("Per-evaluation SingleDetectorLogLikelihoodModelViaArray *3 : ", cost_dict['lnLModelArray']*3)
print("  ... sum of previous two should equal FactoredLogLikelihood. ")
print("Per-evaluation : generate all Ylms : ", cost_dict['Ylms'])
print("Per-evaluation : evaluate one rholm * (2Lmax+1) *3: ", cost_dict['rholms']* (2*Lmax+1)*3)
print("  ... this should agree with SingleDetectorLogLikelihoodData  *3 ")
print(" Special test : polarization log likelihood cost per evaluation = ", cost_dict['psi'],  " which is significantly higher than ", cost_dict['lnL'], "because of one (python-implemented) integration, but should be the same order of time if the integral is done in C ")
#print "Per evaluation NetworkLogLikelihoodTimeMarginalizedDiscrete (an alternative) ", cost_dict['lnLmargTimeDiscrete'], " with answer ", lnLmargTimeD, " which had better compare with the unmarginalized lnL (i.e., about 200) ", lnL
#print "    ... use if convergence of a blind Monte Carlo requires many times the following number of iterations ",  cost_dict['lnLmargTimeDiscrete']/cost_dict['lnL']
print("Per evaluation NetworkLogLikelihoodTimeMarginalized (an alternative) ", cost_dict['lnLmargTime'], " with answer ", lnLmargTime, " which had better compare with the unmarginalized lnL (i.e., about 200) ", lnL)
print("    ... use if convergence of a blind Monte Carlo requires many times the following number of iterations ",  cost_dict['lnLmargTime']/cost_dict['lnL'])




TestDictionary = factored_likelihood_test.TestDictionaryDefault
TestDictionary["UVReport"]              =  analytic_signal  # this report is very confusing for real data
TestDictionary["QSquaredTimeseries"] = False # opts.plot_ShowLikelihoodVersusTime    # should be command-line option to control this plot specifically
TestDictionary["Rho22Timeseries"]      = opts.plot_ShowLikelihoodVersusTime and opts.verbose
TestDictionary["lnLModelAtKnown"]  =  analytic_signal  # this report is very confusing for real data
TestDictionary["lnLDataAtKnownPlusOptimalTimePhase"] = False
TestDictionary["lnLAtKnown"]           = True
TestDictionary["lnLAtKnownMarginalizeTime"]  = False
TestDictionary["AbsolutelyNoPlots"]   = not opts.plot_ShowLikelihoodVersusTime and not opts.plot_ShowSamplerInputs and not opts.plot_ShowSampler
TestDictionary["lnLDataPlot"]            = opts.plot_ShowLikelihoodVersusTime    # Plot individual geocentered L_k(t) and total L(t) [interpolated code]; plot discrete-data L_k(t)
TestDictionary["lnLDataPlotVersusPsi"] = opts.plot_ShowLikelihoodVersusTime
TestDictionary["lnLDataPlotVersusPhi"] = opts.plot_ShowLikelihoodVersusTime
TestDictionary["lnLDataPlotVersusPhiPsi"] = opts.plot_ShowLikelihoodVersusTime
factored_likelihood_test.TestLogLikelihoodInfrastructure(TestDictionary,theEpochFiducial, epoch_post, data_dict, psd_dict, analyticPSD_Q, Psig, rholms,rholms_intp, crossTerms, lookupNK, lookupKN,  ctUArrayDict,ctVArrayDict, rholmArrayDict, detectors,Lmax,opts)




# mcsamplerAdaptiveVolume
#
# Algorithm: based on Tiwari VARAHA https://arxiv.org/pdf/2303.01463.pdf
# Based strongly on 'varaha_example.ipynb' email from 2023/03/07


import sys
import math
#import bisect
from collections import defaultdict

import numpy
np=numpy #import numpy as np
from scipy import integrate, interpolate, special
import itertools
import functools

import os

try:
  import cupy
  import cupyx   # needed for logsumexp
  xpy_default=cupy
  try:
    xpy_special_default = cupyx.scipy.special
    if not(hasattr(xpy_special_default,'logsumexp')):
          print(" mcsamplerAV: no cupyx.scipy.special.logsumexp, fallback mode ...")
          xpy_special_default= special
  except:
    print(" mcsamplerAV: no cupyx.scipy.special, fallback mode ...")
    xpy_special_default= special
  identity_convert = cupy.asnumpy
  identity_convert_togpu = cupy.asarray
  junk_to_check_installed = cupy.array(5)  # this will fail if GPU not installed correctly
  cupy_ok = True
  cupy_pi = cupy.array(np.pi)

  from RIFT.interpolators.interp_gpu import interp

#  from logging import info as log
#  import inspect
#  def verbose_cupy_asarray(*args, **kwargs):
#     print("Transferring data to VRAM", *args, **kwargs)
#     return cupy.asarray(*args, **kwargs)
#  def verbose_cupy_asnumpy(*args, **kwargs):
#     curframe = inspect.currentframe()
#     calframe = inspect.getouterframes(curframe, 2)
#     log("Transferring data to RAM",calframe[1][3]) #,args[0].__name__) #, *args, **kwargs)
#     return cupy.ndarray.asnumpy(*args, **kwargs)
#  cupy.asarray = verbose_cupy_asarray  
#  cupy.ndarray.asnumpy = verbose_cupy_asnumpy

except:
  print(' no cupy (mcsamplerAV)')
#  import numpy as cupy  # will automatically replace cupy calls with numpy!
  xpy_default=numpy  # just in case, to make replacement clear and to enable override
  xpy_special_default = special
  identity_convert = lambda x: x  # trivial return itself
  identity_convert_togpu = lambda x: x
  cupy_ok = False
  cupy_pi = np.pi

def set_xpy_to_numpy():
   xpy_default=numpy
   identity_convert = lambda x: x  # trivial return itself
   identity_convert_togpu = lambda x: x
   cupy_ok = False
   

if 'PROFILE' not in os.environ:
   def profile(fn):
        return fn

if not( 'RIFT_LOWLATENCY'  in os.environ):
    # Dont support selected external packages in low latency
 try:
    import healpy
 except:
    print(" - No healpy - ")

from RIFT.integrators.statutils import  update,finalize, init_log,update_log,finalize_log

#from multiprocessing import Pool

from RIFT.likelihood import vectorized_general_tools

__author__ = "R. O'Shaughnessy, V. Tiwari"

rosDebugMessages = True

class NanOrInf(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

### V. Tiwari routines

def get_likelihood_threshold(lkl, lkl_thr, nsel, discard_prob,xpy_here=xpy_default):
    """
    Find the likelihood threshold that encolses a probability
    lkl  : array of likelihoods (on bins)
    lkl_thr: scalar cutoff
    nsel : integer, has to do with size of array of likelihoods used to evaluate for next array.
    discard_prob: threshold on CDF to throw away an entire bin.  Should be very small
    """
    
    w = xpy_here.exp(lkl - np.max(lkl))
    npoints = len(w)
    sumw = xpy_here.sum(w)
    prob = w/sumw
    idx = xpy_here.argsort(prob)
    ecdf = xpy_here.cumsum(prob[idx])
    F = xpy_here.linspace(np.min(ecdf), 1., npoints)
    prob_stop_thr = lkl[idx][ecdf >= discard_prob][0]
    
    lkl_stop_thr = xpy_here.flip(np.sort(lkl))
    if len(lkl_stop_thr)>nsel:
        lkl_stop_thr = lkl_stop_thr[nsel]
    else:
        lkl_stop_thr = lkl_stop_thr[-1]
    lkl_thr = min(lkl_stop_thr, prob_stop_thr)

    truncp = xpy_here.sum(w[lkl < lkl_thr]) / sumw
            
    return identity_convert(lkl_thr), identity_convert(truncp)  # send both to CPU as needed

def sample_from_bins(xrange, dx, bu, ninbin, reject_out_of_range=False):
    
        ndim = xrange.shape[0]
        xlo, xhi = xrange.T[0] + dx * bu, xrange.T[0] + dx * (bu+1)
        x = xpy_default.vstack([xpy_default.random.uniform(xlo[kk], xhi[kk], size = (npb, ndim)) for kk, npb in enumerate(ninbin)])
        # remove points that are out of range.  Due to rounding issues etc, the sampler above can generate points out of range!
        # Note this rejection will bias the integral, because volumes are calculated assuming a regular grid. We *should* fix the grid sizes to integers
        if reject_out_of_range:
          for indx in np.arange(len(xrange)):
            indx_ok = xpy_default.where(xpy_default.logical_and(x[:,indx] >= xrange[indx,0], x[:,indx] <= xrange[indx,1], ))
            x = x[indx_ok]
        return x


class MCSampler(object):
    """
    Class to define a set of parameter names, limits, and probability densities.
    """

    @staticmethod
    def match_params_from_args(args, params):
        """
        Given two unordered sets of parameters, one a set of all "basic" elements (strings) possible, and one a set of elements both "basic" strings and "combined" (basic strings in tuples), determine whether the sets are equivalent if no basic element is repeated.
        e.g. set A ?= set B
        ("a", "b", "c") ?= ("a", "b", "c") ==> True
        (("a", "b", "c")) ?= ("a", "b", "c") ==> True
        (("a", "b"), "d")) ?= ("a", "b", "c") ==> False  # basic element 'd' not in set B
        (("a", "b"), "d")) ?= ("a", "b", "d", "c") ==> False  # not all elements in set B represented in set A
        """
        not_common = set(args) ^ set(params)
        if len(not_common) == 0:
            # All params match
            return True
        if all([not isinstance(i, tuple) for i in not_common]):
            # The only way this is possible is if there are
            # no extraneous params in args
            return False

        to_match, against = [i for i in not_common if not isinstance(i, tuple)], [i for i in not_common if isinstance(i, tuple)]

        matched = []
        import itertools
        for i in range(2, max(list(map(len, against)))+1):
            matched.extend([t for t in itertools.permutations(to_match, i) if t in against])
        return (set(matched) ^ set(against)) == set()


    def __init__(self,n_chunk=400000,**kwargs):
        # Total number of samples drawn
        self.ntotal = 0
        # Parameter names
        self.params = set()
        self.params_ordered = []  # keep them in order. Important to break likelihood function need for names
        self.params_pinned_vals = {}
        # If the pdfs aren't normalized, this will hold the normalization 
        # Cache for the sampling points
        self._rvs = {}
        # parameter -> cdf^{-1} function object
        # params for left and right limits
        self.llim, self.rlim = {}, {}


        self.n_chunk = n_chunk
        self.nbins = None
        self.ninbin = None
        self.adaptive =[]

        self.pdf = {} # not used

        # MEASURES (=priors): ROS needs these at the sampler level, to clearly separate their effects
        # ASSUMES the user insures they are normalized
        self.prior_pdf = {}

        # histogram setup
        self.xpy = numpy
        self.identity_convert = lambda x: x  # if needed, convert to numpy format  (e.g, cupy.asnumpy)

        # sampling tool
        self.V=None  # fractional volume
        self.delta_V=None  # fractional volume
        

    def setup(self, **kwargs):
        ndim = len(self.params)
        self.nbins = np.ones(ndim)
        self.d_adaptive = len(self.adaptive)
        self.indx_adaptive = [self.params_ordered.index(name) for name in self.adaptive]
        self.indx_not_adaptive = list(set(list( range(ndim))) -set( self.indx_adaptive))
        self.binunique = np.array([ndim* [0]])
        self.ninbin   = [self.n_chunk]
        self.my_ranges =  np.array([[self.llim[x],self.rlim[x]] for x in self.params_ordered])
        self.dx = np.diff(self.my_ranges, axis = 1).flatten()  # weird way to code this
        self.dx0  = np.array(self.dx)   # Save initial prior widths (used for initial prior ragne at end/volume)
        self.cycle = 1

        self.V=1
        self.V_s = np.prod([ self.rlim[x] - self.llim[x] for x in self.llim])  # global sampling volume
        self.lnL_thresh = -np.inf
        self.enc_prob = 0.999

        self.is_varaha=True

    def clear(self):
        """
        Clear out the parameters and their settings, as well as clear the sample cache.
        """
        self.params = set()
        self.params_ordered = []
        self.pdf = {}
        self._pdf_norm = defaultdict(lambda: 1.0)
        self._rvs = {}
        self.llim = {}
        self.rlim = {}
        self.adaptive = []


    def add_parameter(self, params, pdf,  cdf_inv=None, left_limit=None, right_limit=None, prior_pdf=None, adaptive_sampling=False):
        """
        Add one (or more) parameters to sample dimensions. params is either a string describing the parameter, or a tuple of strings. The tuple will indicate to the sampler that these parameters must be sampled together. left_limit and right_limit are on the infinite interval by default, but can and probably should be specified. If several params are given, left_limit, and right_limit must be a set of tuples with corresponding length. Sampling PDF is required, and if not provided, the cdf inverse function will be determined numerically from the sampling PDF.
        """
        self.params.add(params) # does NOT preserve order in which parameters are provided
        self.params_ordered.append(params)
        if rosDebugMessages: 
            print(" Adding parameter ", params, " with limits ", [left_limit, right_limit])
        if isinstance(params, tuple):
            assert all([lim[0] < lim[1] for lim in zip(left_limit, right_limit)])
            if left_limit is None:
                self.llim[params] = list(float("-inf"))*len(params)
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = list(float("+inf"))*len(params)
            else:
                self.rlim[params] = right_limit
        else:
            assert left_limit < right_limit
            if left_limit is None:
                self.llim[params] = float("-inf")
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = float("+inf")
            else:
                self.rlim[params] = right_limit
        self.pdf[params] = pdf
        self.prior_pdf[params] = prior_pdf

        if adaptive_sampling:
            print("   Adapting ", params)
            self.adaptive.append(params)

    def prior_prod(self, x):
        """
        Evaluates prior_pdf(x), multiplying together all factors
        """
        p_out = xpy_default.ones(len(x))
        indx = 0
        for param in self.params_ordered:
            p_out *= self.prior_pdf[param](x[:,indx])
            indx +=1
        return p_out


    def draw_simplified(self,n_to_get, *args, **kwargs):
        rv, log_p = self.draw_simple()
        p = np.exp(log_p)[:n_to_get]
        ps = self.xpy.ones(len(p))*self.V_s/self.V   # sampling prior, full hypercube normalized to 1
        ps = ps[:n_to_get]
        rv = rv[:n_to_get].T
        return ps, p, rv

    def draw_simple(self):
        # Draws
        x =  sample_from_bins(self.my_ranges, self.dx, self.binunique, self.ninbin)
        # if pinning, assign hard values. Note this means prior probabilities are still propagated as arbitrary scales
        if self.params_pinned_vals:
            for p in self.params_pinned_vals:
               indx_p = self.params_ordered.index(p)
               x[:,indx_p] = self.params_pinned_vals[p]
          
        # probabilities at these points.  
        log_p = np.log(self.prior_prod(x))
        # Not including any sampling prior factors, since it is de facto uniform right now (just discarding 'irrelevant' regions)
        return x, log_p

    def update_sampling_prior_selfish(self, lnF, *args, xpy=xpy_default,no_protect_names=True,**kwargs):
        """
      update_sampling_prior

      Update VARAHA sampling hypercubes/

      Note that external samples are NOT uniform.
      VARAHA should only be trained on its own samples, not others!

      We therefore do a single pure step of VARAHA, including *independent* draws.  We will keep state about 'V' etc from previous iterations.
        We therefore also have to know about the function we are integrating. However, we do not keep track of the integral result here -- the top -level routine does this.
       """
        xpy_here = self.xpy
        enforce_bounds=True

        # VT specific items
        loglkl_thr = -1e15
        enc_prob = 0.999 #The approximate upper limit on the final probability enclosed by histograms.
        V = self.V  # nominal scale factor for hypercube volume
        ndim = len(self.params_ordered)
        allx, allloglkl = np.transpose([[]] * ndim), []
        allp = []
        trunc_p = 1e-10 #How much probability analysis removes with evolution
        nsel = 1000# number of largest log-likelihood samples selected to estimate lkl_thr for the next cycle.
        if cupy_ok:
          allx = identity_convert_togpu(allx)
          allloglkl = identity_convert_togpu(allloglkl)

        ntotal_true = 0
        if True: # while (eff_samp < neff and ntotal_true < nmax ): #  and (not bConvergenceTests):
            # Draw samples. Note state variables binunique, ninbin -- so we can re-use the sampler later outside the loop
            rv, log_joint_p_prior = self.draw_simple()  # Beware reversed order of rv
            ntotal_true += len(rv)
            if cupy_ok:
              rv = identity_convert_togpu(rv) # send random numbers to GPU : ugh
              log_joint_p_prior = identity_convert_togpu(log_joint_p_prior)    # send to GPU if required. Don't waste memory reassignment otherwise

            # Evaluate function, protecting argument order
            if True: #'no_protect_names' in kwargs:
                unpacked0 = rv.T
                lnL = lnF(*unpacked0)  # do not protect order
            # else:
            #     unpacked = dict(list(zip(self.params_ordered,rv.T)))
            #     lnL= lnF(**unpacked)  # protect order using dictionary
            # take log if we are NOT using lnL
            if cupy_ok:
              if not(isinstance(lnL,cupy.ndarray)):
                lnL = identity_convert_togpu(lnL)  # send to GPU, if not already there


            # For now: no prior, just duplicate VT algorithm
            log_integrand =lnL  + log_joint_p_prior
            
            loglkl = log_integrand # note we are putting the prior in here

            idxsel = xpy_here.where(loglkl > loglkl_thr)
            #only admit samples that lie inside the live volume, i.e. one that cross likelihood threshold
            allx = xpy_here.append(allx, rv[idxsel], axis = 0)
            allloglkl = xpy_here.append(allloglkl, loglkl[idxsel])
            allp = xpy_here.append(allp, log_joint_p_prior[idxsel])
            ninj = len(allloglkl)


            #just some test to verify if we dont discard more than 1 - Pthr probability
            at_final_threshold = np.round(enc_prob/trunc_p) - np.round(enc_prob/(1 - enc_prob)) == 0
            #Estimate likelihood threshold
            if not(at_final_threshold):
                loglkl_thr, truncp = get_likelihood_threshold(allloglkl, loglkl_thr, nsel, 1 - enc_prob - trunc_p,xpy_here=xpy_here)
                trunc_p += truncp
    
            # Select with threshold
            idxsel = xpy_here.where(allloglkl > loglkl_thr)
            allloglkl = allloglkl[idxsel]
            allp = allp[idxsel]
            allx = allx[idxsel]
            nrec = len(allloglkl)   # recovered size of active volume at present, after selection

            # Weights
            lw = allloglkl - xpy_here.max(allloglkl)
            w = xpy_here.exp(lw)
            neff_varaha = identity_convert(xpy_here.sum(w) ** 2 / xpy_here.sum(w ** 2))
            eff_samp = identity_convert(xpy_here.sum(w)/xpy_here.max(w))  # to CPU as needed
 
            #New live volume based on new likelihood threshold
            V *= (nrec / ninj)
            delta_V = V / np.sqrt(nrec) 
 
            # Redefine bin sizes, reassign points to redefined hypercube set. [Asymptotically this becomes stationary]
            # Note hypercube calculation is on CPU at present, always
            if self.d_adaptive > 0:
              self.nbins = np.ones(ndim)*(1/delta_V) ** (1/self.d_adaptive)  # uniform split in each dimension is normal, but we have array - can be irregular
              self.nbins[self.indx_not_adaptive] = 1  # reset to 1 bin for non-adaptive dimensions
            else:
              self.nbins = np.ones(ndim) # why are we even doing this!

            # bin sizes integers?  May slow us down
            if enforce_bounds:
              self.nbins = np.floor(self.nbins)

            self.dx = np.diff(self.my_ranges, axis = 1).flatten() / self.nbins   # update bin widths
            binidx = ( (( identity_convert(allx) - self.my_ranges.T[0]) / self.dx.T).astype(int)  ) #bin indexs of the samples ... sent back to CPU as needed

            self.binunique = np.unique(binidx, axis = 0)
            self.ninbin = ((self.n_chunk // self.binunique.shape[0] + 1) * np.ones(self.binunique.shape[0])).astype(int)

            self.cycle += 1

        self.V = V
        self.delta_V  = delta_V
        

    @profile
    def integrate_log(self, lnF, *args, xpy=xpy_default,**kwargs):
        """
        Integrate exp(lnF) returning lnI, by using n sample points, assuming integrand is lnF
        Does NOT allow for tuples of arguments, an unused feature in mcsampler

        tempering is done with lnF, suitably modified.

        kwargs:
        nmax -- total allowed number of sample points, will throw a warning if this number is reached before neff.
        neff -- Effective samples to collect before terminating. If not given, assume infinity
        n -- Number of samples to integrate in a 'chunk' -- default is 1000
        save_integrand -- Save the evaluated value of the integrand at the sample points with the sample point
        history_mult -- Number of chunks (of size n) to use in the adaptive histogramming: only useful if there are parameters with adaptation enabled
        tempering_exp -- Exponent to raise the weights of the 1-D marginalized histograms for adaptive sampling prior generation, by default it is 0 which will turn off adaptive sampling regardless of other settings
        temper_log -- Adapt in min(ln L, 10^(-5))^tempering_exp
        tempering_adapt -- Gradually evolve the tempering_exp based on previous history.
        floor_level -- *total probability* of a uniform distribution, averaged with the weighted sampled distribution, to generate a new sampled distribution
        n_adapt -- number of chunks over which to allow the pdf to adapt. Default is zero, which will turn off adaptive sampling regardless of other settings
        convergence_tests - dictionary of function pointers, each accepting self._rvs and self.params as arguments. CURRENTLY ONLY USED FOR REPORTING
        Pinning a value: By specifying a kwarg with the same of an existing parameter, it is possible to "pin" it. The sample draws will always be that value, and the sampling prior will use a delta function at that value.
        """


        xpy_here = self.xpy

        #
        # Pin values
        #
        for p, val in list(kwargs.items()):
            reset_indexes = False
            if p in self.params_ordered:
              reset_indexes = True
              # add to list of pinned values
              self.params_pinned_vals[p] = val
              # disable adaptivity in this parameter, if present
              if p in self.adaptive:
                self.adaptive.remove(p)
            if reset_indexes:
              ndim = len(self.params)
              self.indx_adaptive = [self.params_ordered.index(name) for name in self.adaptive]
              self.indx_not_adaptive = list(set(list( range(ndim))) -set( self.indx_adaptive))

        
        #
        # Determine stopping conditions
        #
        nmax = kwargs["nmax"] if "nmax" in kwargs else float("inf")
        neff = kwargs["neff"] if "neff" in kwargs else numpy.float128("inf")
        n = int(kwargs["n"] if "n" in kwargs else min(100000, nmax))
        convergence_tests = kwargs["convergence_tests"] if "convergence_tests" in kwargs else None
        save_no_samples = kwargs["save_no_samples"] if "save_no_samples" in kwargs else None


        #
        # Adaptive sampling parameters
        #
        n_history = int(kwargs["history_mult"]*n) if "history_mult" in kwargs else 2*n
        if n_history<=0:
            print("  Note: cannot adapt, no history ")

        tempering_exp = kwargs["tempering_exp"] if "tempering_exp" in kwargs else 0.0
        n_adapt = int(kwargs["n_adapt"]*n) if "n_adapt" in kwargs else 1000  # default to adapt to 1000 chunks, then freeze
        floor_integrated_probability = kwargs["floor_level"] if "floor_level" in kwargs else 0
        temper_log = kwargs["tempering_log"] if "tempering_log" in kwargs else False
        tempering_adapt = kwargs["tempering_adapt"] if "tempering_adapt" in kwargs else False
            

        save_intg = kwargs["save_intg"] if "save_intg" in kwargs else False
        # FIXME: The adaptive step relies on the _rvs cache, so this has to be
        # on in order to work
        if n_adapt > 0 and tempering_exp > 0.0:
            save_intg = True

        deltalnL = kwargs['igrand_threshold_deltalnL'] if 'igrand_threshold_deltalnL' in kwargs else float("Inf") # default is to return all
        deltaP    = kwargs["igrand_threshold_p"] if 'igrand_threshold_p' in kwargs else 0 # default is to omit 1e-7 of probability
        bFairdraw  = kwargs["igrand_fairdraw_samples"] if "igrand_fairdraw_samples" in kwargs else False
        n_extr = kwargs["igrand_fairdraw_samples_max"] if "igrand_fairdraw_samples_max" in kwargs else None

        bShowEvaluationLog = kwargs['verbose'] if 'verbose' in kwargs else False
        bShowEveryEvaluation = kwargs['extremely_verbose'] if 'extremely_verbose' in kwargs else False


        verbose = kwargs["verbose"] if "verbose" in kwargs else False  # default
        super_verbose = kwargs["super_verbose"] if "super_verbose" in kwargs else False  # default
        dict_return_q = kwargs["dict_return"] if "dict_return" in kwargs else False  # default.  Method for passing back rich data structures for debugging

        # use integer numbers of bins always
        enforce_bounds = kwargs["enforce_bounds"] if "enforce_bounds" in kwargs else False

        if bShowEvaluationLog:
            print(" .... mcsampler : providing verbose output ..... ")

        current_log_aggregate = None
        eff_samp = 0  # ratio of max weight to sum of weights
        maxlnL = -np.inf  # max lnL
        maxval=0   # max weight
        outvals=None  # define in top level scope
        self.ntotal = 0
        if bShowEvaluationLog:
            print("iteration Neff  sqrt(2*lnLmax) sqrt(2*lnLmarg) ln(Z/Lmax) int_var")

        self.n_chunk = n
        self.setup()  # sets up self.my_ranges, self.dx initially

        cycle =1

        # VT specific items
        loglkl_thr = -1e15
        enc_prob = 0.999 #The approximate upper limit on the final probability enclosed by histograms.
        V = 1  # nominal scale factor for hypercube volume
        ndim = len(self.params_ordered)
        allx, allloglkl = np.transpose([[]] * ndim), []
        allp = []
        trunc_p = 1e-10 #How much probability analysis removes with evolution
        nsel = 1000# number of largest log-likelihood samples selected to estimate lkl_thr for the next cycle.
        nsel = np.min([nsel, int(0.1*self.n_chunk)]) #  if chunk size is small, don't pick too many points
        if cupy_ok:
          allx = identity_convert_togpu(allx)
          allloglkl = identity_convert_togpu(allloglkl)

        ntotal_true = 0
        while (eff_samp < neff and ntotal_true < nmax ): #  and (not bConvergenceTests):
            # Draw samples. Note state variables binunique, ninbin -- so we can re-use the sampler later outside the loop
            rv, log_joint_p_prior = self.draw_simple()  # Beware reversed order of rv
            ntotal_true += len(rv)
            if cupy_ok:
              rv = identity_convert_togpu(rv) # send random numbers to GPU : ugh
              log_joint_p_prior = identity_convert_togpu(log_joint_p_prior)    # send to GPU if required. Don't waste memory reassignment otherwise

            # Evaluate function, protecting argument order
            if 'no_protect_names' in kwargs:
                unpacked0 = rv.T
                lnL = lnF(*unpacked0)  # do not protect order
            else:
                unpacked = dict(list(zip(self.params_ordered,rv.T)))
                lnL= lnF(**unpacked)  # protect order using dictionary
            # take log if we are NOT using lnL
            if cupy_ok:
              if not(isinstance(lnL,cupy.ndarray)):
                lnL = identity_convert_togpu(lnL)  # send to GPU, if not already there


            # For now: no prior, just duplicate VT algorithm
            log_integrand =lnL  + log_joint_p_prior
#            log_weights = tempering_exp*lnL + log_joint_p_prior
            # log aggregate: NOT USED at present, remember the threshold is floating
            if current_log_aggregate is None:
              current_log_aggregate = init_log(log_integrand,xpy=xpy,special=xpy_special_default)
            else:
              current_log_aggregate = update_log(current_log_aggregate, log_integrand,xpy=xpy,special=xpy_special_default)
            
            loglkl = log_integrand # note we are putting the prior in here

            idxsel = xpy_here.where(loglkl > loglkl_thr)
            #only admit samples that lie inside the live volume, i.e. one that cross likelihood threshold
            allx = xpy_here.append(allx, rv[idxsel], axis = 0)
            allloglkl = xpy_here.append(allloglkl, loglkl[idxsel])
            allp = xpy_here.append(allp, log_joint_p_prior[idxsel])
            ninj = len(allloglkl)


            #just some test to verify if we dont discard more than 1 - Pthr probability
            at_final_threshold = np.round(enc_prob/trunc_p) - np.round(enc_prob/(1 - enc_prob)) == 0
            #Estimate likelihood threshold
            if not(at_final_threshold):
                loglkl_thr, truncp = get_likelihood_threshold(allloglkl, loglkl_thr, nsel, 1 - enc_prob - trunc_p,xpy_here=xpy_here)
                trunc_p += truncp
    
            # Select with threshold
            idxsel = xpy_here.where(allloglkl > loglkl_thr)
            allloglkl = allloglkl[idxsel]
            allp = allp[idxsel]
            allx = allx[idxsel]
            nrec = len(allloglkl)   # recovered size of active volume at present, after selection

            # Weights
            lw = allloglkl - xpy_here.max(allloglkl)
            w = xpy_here.exp(lw)
            neff_varaha = identity_convert(xpy_here.sum(w) ** 2 / xpy_here.sum(w ** 2))
            eff_samp = identity_convert(xpy_here.sum(w)/xpy_here.max(w))  # to CPU as needed
 
            #New live volume based on new likelihood threshold
            V *= (nrec / ninj)
            delta_V = V / np.sqrt(nrec) 
 
            # Redefine bin sizes, reassign points to redefined hypercube set. [Asymptotically this becomes stationary]
            # Note hypercube calculation is on CPU at present, always
            if self.d_adaptive > 0:
              self.nbins = np.ones(ndim)*(1/delta_V) ** (1/self.d_adaptive)  # uniform split in each dimension is normal, but we have array - can be irregular
              self.nbins[self.indx_not_adaptive] = 1  # reset to 1 bin for non-adaptive dimensions
            else:
              self.nbins = np.ones(ndim) # why are we even doing this!

            # bin sizes integers?  May slow us down
            if enforce_bounds:
              self.nbins = np.floor(self.nbins)

            self.dx = np.diff(self.my_ranges, axis = 1).flatten() / self.nbins   # update bin widths
            binidx = ( (( identity_convert(allx) - self.my_ranges.T[0]) / self.dx.T).astype(int)  ) #bin indexs of the samples ... sent back to CPU as needed

            self.binunique = np.unique(binidx, axis = 0)
            self.ninbin = ((self.n_chunk // self.binunique.shape[0] + 1) * np.ones(self.binunique.shape[0])).astype(int)
            self.ntotal = current_log_aggregate[0]

            if super_verbose:
              print(ntotal_true,eff_samp, np.round(neff_varaha), np.round(np.max(allloglkl), 1), len(allloglkl), np.mean(self.nbins), V,  len(self.binunique),  np.round(loglkl_thr, 1), trunc_p)
            else:
              print(ntotal_true,eff_samp, np.sqrt(2*xpy_here.max(allloglkl - allp)), '-', np.log(V), np.sqrt(xpy_here.var(w/xpy_here.mean(w))/len(w) ))

            cycle += 1
            if cycle > 1000:
                break

        # VT approach was to accumulate samples, but then prune them.  So we have all the lnL and x draws

        # write in variables requested in the standard format
        for indx in np.arange(len(self.params_ordered)):
            self._rvs[self.params_ordered[indx]] = allx[:,indx]  # pull out variable
        # write out log integrand
        self._rvs['log_integrand']  = allloglkl - allp  # remember 'allloglkl' really is Lp -- despite the misleading name! --  so we are *undoing* that
        self._rvs['log_joint_prior'] = allp
        self._rvs['log_joint_s_prior'] = xpy_here.ones(len(allloglkl))*(np.log(1/V) - np.sum(np.log(self.dx0)))  # effective uniform sampling on this volume

        # Manual estimate of integrand, done transparently (no 'log aggregate' or running calculation -- so memory hog
        log_wt = self._rvs["log_integrand"] + self._rvs["log_joint_prior"] - self._rvs["log_joint_s_prior"]
        log_wt = identity_convert(log_wt)  # convert to CPU
        log_int = special.logsumexp( log_wt) - np.log(len(log_wt))  # mean value
        rel_var = np.var( np.exp(log_wt - log_int))/len(log_wt)   # error in integral, estimated: just taking int = <w> , so error is V(w_k)/N (sample mean/variance)
        eff_samp = np.sum(np.exp(log_wt - np.max(log_wt)))
        maxval = np.max(allloglkl)  # max of log

        # Integral value: NOT RELIABLE b/c not just using samples in 
#        outvals = finalize_log(current_log_aggregate,xpy=xpy)
#        log_wt_tmp = allloglkl[np.isfinite(allloglkl)]  # remove infinite entries
#        outvals = init_log(log_wt_tmp)
#        print(outvals, log_int, maxval, current_log_aggregate)
#        eff_samp = xpy.exp(  outvals[0]+np.log(len(allloglkl)) - maxval)   # integral value minus floating point, which is maximum
#        rel_var = np.exp(outvals[1]/2  - outvals[0]  - np.log(self.ntotal)/2 )

        # Do a fair draw of points, if option is set. CAST POINTS BACK TO NUMPY, IDEALLY
        if bFairdraw and not(n_extr is None):
           n_extr = int(numpy.min([n_extr,1.5*identity_convert(eff_samp),1.5*neff]))
           print(" Fairdraw size : ", n_extr)
           ln_wt = self.xpy.array(self._rvs["log_integrand"] + self._rvs["log_joint_prior"] - self._rvs["log_joint_s_prior"] ,dtype=float)
           ln_wt = identity_convert(ln_wt)  # send to CPU
           ln_wt += - special.logsumexp(ln_wt)
           wt = xpy.exp(identity_convert_togpu(ln_wt))
           if n_extr < len(self._rvs["log_integrand"]):
               indx_list = self.xpy.random.choice(self.xpy.arange(len(wt)), size=n_extr,replace=True,p=wt) # fair draw
               # FIXME: See previous FIXME
               for key in list(self._rvs.keys()):
                   if isinstance(key, tuple):
                       self._rvs[key] = identity_convert(self._rvs[key][:,indx_list])
                   else:
                       self._rvs[key] = identity_convert(self._rvs[key][indx_list])


        # perform type conversion of all stored variables.  VERY LARGE -- should only do this if we need it!
        if cupy_ok:
          for name in self._rvs:
            if isinstance(self._rvs[name],xpy_default.ndarray):
              self._rvs[name] = identity_convert(self._rvs[name])   # this is trivial if xpy_default is numpy, and a conversion otherwise

        dict_return = {}
        return log_int, np.log(rel_var)  +2*log_int, eff_samp, dict_return

        # if outvals:
        #   out0 = outvals[0]; out1 = outvals[1]
        #   if not(isinstance(outvals[0], np.float64)):
        #     # type convert everything as needed
        #     out0 = identity_convert(out0)
        #   if not(isinstance(outvals[1], np.float64)):
        #     out1 = identity_convert(out1)
        #     eff_samp = identity_convert(eff_samp)
        #   return out0, out1 - np.log(self.ntotal), eff_samp, dict_return
        # else: # very strange case where we terminate early
        #   return None, None, None, None


    @profile
    def integrate(self, func, *args, **kwargs):
        """
        Integrate func, by using n sample points. Right now, all params defined must be passed to args must be provided, but this will change soon.
        Does NOT allow for tuples of arguments, an unused feature in mcsampler

        kwargs:
        nmax -- total allowed number of sample points, will throw a warning if this number is reached before neff.
        neff -- Effective samples to collect before terminating. If not given, assume infinity
        n -- Number of samples to integrate in a 'chunk' -- default is 1000
        save_integrand -- Save the evaluated value of the integrand at the sample points with the sample point
        history_mult -- Number of chunks (of size n) to use in the adaptive histogramming: only useful if there are parameters with adaptation enabled
        tempering_exp -- Exponent to raise the weights of the 1-D marginalized histograms for adaptive sampling prior generation, by default it is 0 which will turn off adaptive sampling regardless of other settings
        temper_log -- Adapt in min(ln L, 10^(-5))^tempering_exp
        tempering_adapt -- Gradually evolve the tempering_exp based on previous history.
        floor_level -- *total probability* of a uniform distribution, averaged with the weighted sampled distribution, to generate a new sampled distribution
        n_adapt -- number of chunks over which to allow the pdf to adapt. Default is zero, which will turn off adaptive sampling regardless of other settings
        convergence_tests - dictionary of function pointers, each accepting self._rvs and self.params as arguments. CURRENTLY ONLY USED FOR REPORTING
        Pinning a value: By specifying a kwarg with the same of an existing parameter, it is possible to "pin" it. The sample draws will always be that value, and the sampling prior will use a delta function at that value.
        """
        def ln_func(*args):
          return np.log(func(*args))
        infunc = ln_func
        use_lnL=False
        if 'use_lnL' in kwargs:   # should always be positive
          if kwargs['use_lnL']:
            infunc = func
            use_lnL=True
        log_int_val, log_var, eff_samp, dict_return =  self.integrate_log(func, **kwargs)  # pass it on, easier than mixed coding
        if use_lnL:
          self._rvs['integrand'] = self._rvs["log_integrand"]

        return log_int_val, log_var, eff_samp, dict_return


# PROVIDE CROSS-CODE DEPLOY, without rewrite
from RIFT.integrators.mcsamplerGPU import  uniform_samp_cdf_inv_vector,     ret_uniform_samp_vector_alt,uniform_samp_phase, cos_samp_vector, cos_samp_cdf_inv_vector,ret_uniform_samp_vector_alt, uniform_samp_theta,uniform_samp_psi, dec_samp_vector, dec_samp_cdf_inv_vector, uniform_samp_dec
from RIFT.integrators.mcsamplerGPU import q_samp_vector, M_samp_vector,q_cdf_inv_vector



#! /usr/bin/env python
# Example
#   /scripts/localize_ini.py --ifo-classifier review_test_data/4s/all_networks.json --event 0 --ini proto_4s.ini --sim-xml scripts/mdc.xml.gz --mc-file-settings here_mc_ranges --guess-snr 20 --duration 4
import json
import sys
import argparse
import RIFT.lalsimutils as lalsimutils
import numpy as np
import configparser

parser = argparse.ArgumentParser()
#parser.add_argument("--ifos",help="Space-separated list of IFO names, eg 'H1 L1 V1', to use for this event ")
parser.add_argument("--event",type=int)
parser.add_argument("--ini")
parser.add_argument("--sim-xml")
parser.add_argument("--guess-snr",type=float)
parser.add_argument("--mc-range")
parser.add_argument("--cip-sampler",type=str)
parser.add_argument("--ile-sampler",type=str)
parser.add_argument("--assume-nonprecessing",action='store_true')
parser.add_argument("--assume-fixed-sky-location",action='store_true')
#parser.add_argument("--mc-range-limits-snr-safety",type=float,default=8)
#parser.add_argument("--duration",type=int)
opts = parser.parse_args()

#
#ifos =None
print("Localizing ini file for event {}".format(opts.event))
# Get IFOs for it
#with open(opts.ifos,'r') as f:
#    dat_ifos = json.load(f)
#    ifos  = dat_ifos[opts.event] 

# Get event time from saved event object
# sim-xml only has one event remember
P=lalsimutils.xml_to_ChooseWaveformParams_array(opts.sim_xml)[opts.event]

# open mc file
#line = None
#mc_range=eval(opts.mc_range.replace(' ', '')) # completely eliminate whitespace

snr_cut = 30
snr_cut= float(snr_cut)

#########################
# Open defaults ini files
#########################
config=configparser.ConfigParser()
config.optionxform = str
config.read('pseudo_ini.ini')

#########################
# Change default settings
# if needed for current run
#########################
def modify_ini(config, section, option, new_val):
    # Check and modify the option if needed
    if (not config.has_option(section, option)) or (config.get(section, option) != new_val):
        config[section][option] = new_val
        print(f"{option} in {section} changed.")
        return True
    else:
        #print(f"{option} in {section} unchanged.")
        return False

changes_made = False
with open(opts.ini,'r') as f:
    line = f.readline()
    print_explode = False
    ## make NOT a while loop, just scan file
    for line in f:
        changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'force-hint-snr', str(opts.guess_snr))
        changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'event-time', str(P.tref))
        if 'cip-explode-jobs' in line:
            print(line.rstrip(), print_explode,file=sys.stderr)
            if not(print_explode):
                changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'cip-explode-jobs-auto', 'True')
                print_explode=True
            continue
        elif opts.guess_snr > 37.5:
            changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'use-downscale-early', 'True')
            changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'internal-ile-use-lnL', 'True')
            changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'internal-ile-sky-network-coordinates', 'True')
            changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'cip-sigma-cut', '0.7')
            if opts.guess_snr > 100:
                changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'internal-cip-use-lnL', 'True')
                changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'internal-ile-rotate-phase', 'True')
                changes_made |= modify_ini(config, 'rift-pseudo-pipe', 'ile-sampler-method', "'GMM'")

#########################
# Write new ini file with
# changed settings
#########################
if changes_made:
        with open('localize.ini', 'w') as f:
            config.write(f)
        print("File saved.")
exit(0)


#########################
# old localize
#########################
with open(opts.ini,'r') as f:
    line = f.readline()

    print_explode=False
    while line:
     if line.startswith('ifos='):
        print('ifos={}'.format(ifos))
     elif 'cip-explode-jobs' in line: #.startswith('cip-explode-jobs'):
        print(line.rstrip(), print_explode,file=sys.stderr)
        if not(print_explode):
           print('cip-explode-jobs-auto=True')  # basically always do this, override proto settings which were not well chosen
           print_explode=True
     elif opts.guess_snr > 37.5 and line.startswith('internal-ile-auto-logarithm-offset'):
         True # Don't use the logarithm offset when lnL is active for ILE
     elif line.startswith('internal-ile-sky-network-coordinates'):
        True # skip this line, set with IFO list at end
     else:
        clean_line = line.strip()
        if (clean_line):
          print(clean_line)
     line  = f.readline()

print("force-hint-snr={}".format(opts.guess_snr))
if opts.guess_snr<snr_cut:
    print("force-mc-range={}".format(opts.mc_range))
else:
    print("limit-mc-range={}".format(mc_range))

sky_printed=False
# Try to set most of the high-SNR options here. Distance marginalization will be left to the top-level script
if opts.ile_sampler != 'AV':
  if opts.guess_snr>37.5:
    print("use-downscale-early=True")
    print("internal-ile-use-lnL=True")
    if not(opts.fix_sky_location):
        print("internal-ile-sky-network-coordinates=True")
    print("cip-sigma-cut=0.7")
  if opts.guess_snr>100:
    print("internal-cip-use-lnL=True")
    print("internal-ile-rotate-phase=True")
    print("ile-sampler-method=GMM")
elif opts.ile_sampler == 'AV':
    print("internal-ile-use-lnL=True")  # automatic/redundant
    if opts.guess_SNR > 26:
        print("manual-extra-ile-args=--force-adapt-all")  # adapt in all parameters
#if len(ifos) == 2 and not sky_printed:
#    print("internal-ile-sky-network-coordinates=True")
print("event-time={}".format(P.tref))



#! /usr/bin/env python
# prior_pp_check:  produces PP-like output from injection files, compares to prior.
# Note currently does NOT change the distance prior to be cosmological

import argparse
import RIFT.lalsimutils as lalsimutils
import numpy as np
import glob
import RIFT.likelihood.priors_utils as priors_utils
import RIFT.integrators.mcsampler  as mcsampler # for quick CDF inverse
import functools
import lal

parser = argparse.ArgumentParser()
parser.add_argument("--base-dir",default='my_set_16s',type=str)
parser.add_argument("--mc-range",default='[10,20]',type=str)
parser.add_argument("--eta-range", default='[0.2,0.2499999]',type=str)
parser.add_argument("--d-max",default=750,type=float)
parser.add_argument("--parameter",action='append',type=str)
parser.add_argument("--plot-sanity",action='store_true')
opts = parser.parse_args()
mc_min,mc_max = eval(opts.mc_range)
eta_min,eta_max = eval(opts.eta_range)
q_min = (1-np.sqrt(1-4*eta_min))/(1+np.sqrt(1-4*eta_min))

import scipy.integrate
#def q_prior(x,nm=1):
#    return 1./(1+x)**2/nm
#nm_q = scipy.integrate.quad(q_prior,1./20, 1)[0]
def delta_mc_prior(x,norm_factor=1.44):
    """
    delta_mc = sqrt(1-4eta)  <-> eta = 1/4(1-delta^2)
    Transform the prior above
    """
    eta_here = 0.25*(1 -x*x)
    return 2./np.power(eta_here, 6./5.)/norm_factor


# Distance prior: use mcsampler to make cdf inverse
sampler = mcsampler.MCSampler()
nm = priors_utils.dist_prior_pseudo_cosmo_eval_norm(0,opts.d_max)
dist_prior_pdf =functools.partial( priors_utils.dist_prior_pseudo_cosmo, nm=nm)
sampler.add_parameter('dist', pdf=dist_prior_pdf, cdf_inv= None, left_limit=0.0, right_limit=opts.d_max)
#q_prior_pdf = functools.partial( q_prior, nm=nm_q)
sampler.add_parameter('delta_mc',pdf=delta_mc_prior, left_limit=0, right_limit=np.sqrt(1-4*eta_min))

npts = 10000
# Generate prior data
P_list  =[]
P_out = lalsimutils.ChooseWaveformParams()
for name in ['m1','m2', 's1x','s1y','s1z','s2x', 's2y','s2z', 'theta', 'phi','psi','phiref','dist', 'incl']:
    setattr(P_out,name, np.zeros(npts))

random_mc = np.sqrt(mc_min**2 + np.random.uniform(size=2*npts)*(mc_max**2-mc_min**2))
random_dL = sampler.cdf_inv['dist'](np.random.uniform(size=2*npts))
random_delta = sampler.cdf_inv['delta_mc'](np.random.uniform(size=2*npts))
#print(random_mc,random_dL, random_q)
n_kept = 0
for indx in np.arange(2*npts):
    if n_kept >= npts:
        continue
    P = lalsimutils.ChooseWaveformParams()
    P.randomize() # randomize spins
    P.assign_param('mc', random_mc[indx])
    P.assign_param('delta_mc',random_delta[indx])
    P.assign_param('dist',random_dL[indx])
    if P.m2 > 1 and n_kept < npts:
        P_list.append(P)
        P_out.m1[n_kept] = P.m1
        P_out.m2[n_kept] = P.m2 
        P_out.s1x[n_kept] = P.s1x
        P_out.s1y[n_kept] = P.s1y
        P_out.s1z[n_kept] = P.s1z
        P_out.s2x[n_kept] = P.s2x
        P_out.s2y[n_kept] = P.s2y
        P_out.s2z[n_kept] = P.s2z
        P_out.theta[n_kept] = P.theta
        P_out.phi[n_kept] = P.phi
        P_out.psi[n_kept] = P.psi
        P_out.phiref[n_kept] = P.phiref
        P_out.dist[n_kept] = P.dist
        P_out.incl[n_kept] = P.incl
        n_kept += 1




# Pull in MDC files
fname = 'mdc.xml.gz'
P_list = lalsimutils.xml_to_ChooseWaveformParams_array(fname)
mvals = np.zeros((len(P_list),4))
indx_name =0
for indx, P in enumerate(P_list):
    dat_out =np.zeros(len(opts.parameter))
    if opts.plot_sanity:
        mvals[indx_name] = [P.extract_param('mc')/lal.MSUN_SI,P.extract_param('q'),P.m1/lal.MSUN_SI,P.m2/lal.MSUN_SI]
        if P.m1<P.m2:
            mvals[indx_name] =  [P.extract_param('mc')/lal.MSUN_SI,1./P.extract_param('q'),P.m2/lal.MSUN_SI,P.m1/lal.MSUN_SI]
        indx_name+=1
    indx = 0
    for name in opts.parameter:
        fac =1
        if name in ['mc']:
            fac = lal.MSUN_SI
        if name in ['dist']:
            fac = lal.PC_SI*1e6
        dat_out[indx] = np.sum(P_out.extract_param(name) < P.extract_param(name)/fac)*1.0/npts
        #print(P.extract_param(name),P_out.extract_param(name))
        indx+=1
    print(' '.join(map(str,dat_out)))

if opts.plot_sanity:
    from matplotlib import pyplot as plt
    plt.scatter(mvals[:,0], mvals[:,1])
    plt.savefig("prior_fig_mc_q.png")
    plt.clf()
    plt.scatter(mvals[:,2], mvals[:,3])
    xvals = np.linspace(np.min(mvals[:,3]), np.max(mvals[:,2]), 100)
    yvals = xvals
    plt.plot(xvals,yvals)
    yvals =  q_min*xvals
    plt.plot(xvals, yvals)
    plt.savefig("prior_fig_m1_m2.png")
